import arcgis as _arcgis 
from ._arcgis_model import ArcGISModel
from ..._impl.common._deprecate import deprecated
from .._data import _check_esri_files, _raise_fastai_import_error
import random
import math  
import traceback    
try:
    import pandas
    import tempfile
    import numpy as np
    import json
    import os
    import io
    import shutil
    import warnings
    from pathlib import Path
    from functools import partial
    from ._unet_utils import is_no_color
    from ._codetemplate import feature_classifier_prf
    import torch
    from torchvision import models
    from fastai.metrics import accuracy, MultiLabelFbeta
    from .._utils.metrics import accuracy_multi
    from fastai.vision.image import open_image
    from fastai.data_block import MultiCategoryList
    from fastai.vision.data import ImageDataBunch, ImageList
    from fastai.vision import imagenet_stats, normalize
    from fastai.basic_train import Learner, LearnerCallback
    from torch.utils.data.sampler import WeightedRandomSampler, BatchSampler
    from fastai.vision.learner import cnn_learner, ClassificationInterpretation, cnn_config
    from ._arcgis_model import _set_multigpu_callback, _resnet_family
    from fastai.vision.transform import crop, rotate, dihedral_affine, brightness, contrast, skew, rand_zoom, get_transforms
    import torch.nn.functional as functional
    import glob
    import time
    import xml.etree.ElementTree as ElementTree
    import PIL.Image
    import PIL.ExifTags
    from torch.nn import Module as NnModule
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path, image_batch_stretcher
    from matplotlib import pyplot as plt
    import copy
    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    class NnModule():
        pass
    HAS_FASTAI = False

HAS_ARCPY = True
try: 
    import arcpy 
except Exception: 
    HAS_ARCPY = False

def _mobilenet_split(m:NnModule): return m[0][0][0], m[1]


def _prediction_function(predictions):
    classes = {}
    max_prediction_value = 0
    max_prediction_class = None
    for prediction in predictions:
        if not classes.get(prediction[0]):
            classes[prediction[0]] = prediction[1]
        else:
            classes[prediction[0]] = classes[prediction[0]] + prediction[1]
        if max_prediction_value < classes[prediction[0]]:
            max_prediction_value = classes[prediction[0]]
            max_prediction_class = prediction[0]

    return max_prediction_class, max_prediction_value


class FeatureClassifier(ArcGISModel):
    """
    Creates an image classifier to classify the area occupied by a
    geographical feature based on the imagery it overlaps with.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    backbone                Optional torchvision model. Backbone CNN model to be used for
                            creating the base of the ``FeatureClassifier``, which
                            is ``resnet34`` by default.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    ---------------------   -------------------------------------------
    mixup                   Optional boolean. If set to True, it creates 
                            new training images by randomly mixing training set images.

                            The default is set to False.
    ---------------------   -------------------------------------------
    oversample              Optional boolean. If set to True, it oversamples unbalanced
                            classes of the dataset during training. Not supported with
                            MultiLabel dataset.
    ---------------------   -------------------------------------------
    backend                 Optional string. Controls the backend framework to be used
                            for this model, which is 'pytorch' by default.

                            valid options are 'pytorch', 'tensorflow'
    =====================   ===========================================

    :returns: `FeatureClassifier` Object
    """

    def __init__(self, data, backbone=None, pretrained_path=None, mixup=False, oversample=False, backend='pytorch', *args, **kwargs):
        
        self._backend = backend
        if self._backend == 'tensorflow':
            super().__init__(data, None)
            self._intialize_tensorflow(data, backbone, pretrained_path, mixup, kwargs)
        else:
            super().__init__(data, backbone, **kwargs)

            backbone_cut = None
            backbone_split = None

            _backbone = self._backbone
            if hasattr(self, '_orig_backbone'):
                _backbone = self._orig_backbone
                _backbone_meta = cnn_config(self._orig_backbone)
                backbone_cut = _backbone_meta['cut']
                backbone_split = _backbone_meta['split']

            if _backbone == models.mobilenet_v2:
                backbone_cut = -1
                backbone_split = _mobilenet_split

            if not self._check_backbone_support(_backbone):
                raise Exception (f"Enter only compatible backbones from {', '.join(self.supported_backbones)}")

            self._check_dataset_support(self._data)

            self._code = feature_classifier_prf

            if getattr(data, '_dataset_type', "Labeled_Tiles") == 'MultiLabeled_Tiles':
                # ToDo: allow option to change `thresh` parameter by user
                accuracy_multi.__name__ = "accuracy"
                metrics = [accuracy_multi, MultiLabelFbeta()]
            else:
                metrics = accuracy

            self.learn = cnn_learner(data, self._backbone, metrics=metrics, cut=backbone_cut, split_on=backbone_split)
            if oversample:
                self.learn.callbacks.append(OverSamplingCallback(self.learn))
            self._arcgis_init_callback() # make first conv weights learnable
            
            # Add Mixup data augmentation
            if mixup:
                # For mixup to work with multilabel call it with parameter stack_y=False
                stack_y = getattr(data, '_dataset_type', "Labeled_Tiles") == 'Labeled_Tiles'
                self.learn = self.learn.mixup(stack_y=stack_y)

            self.learn.model = self.learn.model.to(self._device)

            _set_multigpu_callback(self)
            if pretrained_path is not None:
                self.load(pretrained_path)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    @property
    def supported_backbones(self):
        """ Supported torchvision backbones for this model. """
        return FeatureClassifier._supported_backbones()

    @staticmethod
    def _supported_backbones():
        return [*_resnet_family, models.mobilenet_v2.__name__]

    @property
    def  supported_datasets(self):
        """ Supported dataset types for this model. """
        return FeatureClassifier._supported_datasets()
    
    @staticmethod
    def _supported_datasets():
        return ['Labeled_Tiles', 'MultiLabeled_Tiles']  

    def show_results(self, rows=5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.
        """
        self._check_requisites()
        import math
        if (rows ** 2) > len(self._data.valid_ds):
            rows = math.floor(math.sqrt(len(self._data.valid_ds)))

        self.learn.show_results(rows=rows, **kwargs)
   
    def _show_results_multispectral(self, rows=5, **kwargs):
        from .._utils.image_classification import IC_show_results
        return_fig = kwargs.get('return_fig', False)
        fig=IC_show_results(
            self,
            nrows=rows,
            **kwargs
        )
        if return_fig:
            fig1, axs = fig
            return fig1

    def predict(self, img_path):
        """
        Runs prediction on an Image. Works with RGB images only.
        
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        image_path              Required. Path to the image file to make the
                                predictions on.
        =====================   ===========================================
        
        :returns: prediction label and confidence        
        """
        
        img = open_image(img_path)
        return self.learn.predict(img)

    def _predict_batch(self, imagetensor_batch):
        predictions = self.learn.model.eval()(imagetensor_batch.to(self._device)).detach().cpu()
        predictions_conf,  predicted_classes = torch.max(predictions, dim=-1)
        predicted_classes = predicted_classes.tolist()
        predictions_conf = (predictions_conf*100).tolist()
        return predicted_classes, predictions_conf

    def _save_confusion_matrix(self, path):
        from IPython.utils import io
        with io.capture_output() as captured:
            self.plot_confusion_matrix()
            plt.savefig(os.path.join(path, 'confusion_matrix.png'))
            plt.close()
        
    @property
    def _model_metrics(self):
        return {}

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_FeatureClassifier"
        _emd_template["ModelType"] = "ObjectClassification"
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISObjectClassifier.py"
        else:
            _emd_template["InferenceFunction"] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISObjectClassifier.py"
        _emd_template["MetaDataMode"] = self._data._dataset_type
        _emd_template["ExtractBands"] = [0, 1, 2]
        _emd_template['CropSizeFixed'] = 1  # hardcoded
        _emd_template['BlackenAroundFeature'] = 0  # hardcoded
        _emd_template['ImageSpaceUsed'] = "MAP_SPACE"
        _emd_template['Classes'] = []
        class_data = {}

        if self._data._dataset_type == 'MultiLabeled_Tiles':
                self._data.class_mapping = {k: v for k, v in enumerate(self._data.classes)}
        inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}

        for i, class_name in enumerate(self._data.classes):
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            class_data["Color"] = color
            _emd_template['Classes'].append(class_data.copy())

        # if getattr(self, '_is_multispectral', False):
        #     _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        #     _emd_template["ModelConfiguration"] = "_FeatureClassifier"
        #     if save_inference_file:
        #         _emd_template["InferenceFunction"] = "ArcGISObjectClassifier.py"
        #     else:
        #         _emd_template["InferenceFunction"] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISObjectClassifier.py"

        return _emd_template

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a Feature classifier from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :returns: `FeatureClassifier` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)
            
        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd['ModelFile'])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd['ModelParameters']
        chip_size = emd["ImageWidth"]

        try:
            class_mapping = {i['Value']: i['Name'] for i in emd['Classes']}
            color_mapping = {i['Value']: i['Color'] for i in emd['Classes']}
        except KeyError:
            class_mapping = {i['ClassValue']: i['ClassName'] for i in emd['Classes']}
            color_mapping = {i['ClassValue']: i['Color'] for i in emd['Classes']}

        if data is None:
            ranges = (0, 1)
            train_tfms = [rotate(degrees=30, p=0.5),
                          crop(size=chip_size, p=1., row_pct=ranges, col_pct=ranges),
                          dihedral_affine(), brightness(change=(0.4, 0.6)), contrast(scale=(0.75, 1.5)),
                          # rand_zoom(scale=(0.75, 1.5))
                          ]
            val_tfms = [crop(size=chip_size, p=1.0, row_pct=0.5, col_pct=0.5)]
            transforms = (train_tfms, val_tfms)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)

                if ("MetaDataMode" in emd) and (emd["MetaDataMode"] == "MultiLabeled_Tiles"):
                    img_list = ImageList([], path=emd_path.parent.parent).split_by_idx([])
                    data = img_list.label_const(0, label_cls=MultiCategoryList, 
                                                classes=list(class_mapping.values())).transform(transforms).databunch().normalize(imagenet_stats)
                    data._dataset_type = 'MultiLabeled_Tiles'
                else:
                    data = ImageDataBunch.single_from_classes(
                        emd_path.parent.parent, sorted(list(class_mapping.values())),
                        ds_tfms=transforms, size=chip_size).normalize(imagenet_stats)

            data.chip_size = chip_size
            data.class_mapping = class_mapping
            data.classes = list(class_mapping.values())
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            data = get_multispectral_data_params_from_emd(data, emd)

        resize_to = emd.get('resize_to')
        data.resize_to = resize_to

        return cls(data, **model_params, pretrained_path=str(model_file))

    def plot_confusion_matrix(self, **kwargs):
        """
        Plots a confusion matrix of the model predictions to evaluate accuracy
        kwargs: 'thresh' - confidence score threshold for multilabel predictions, defaults to 0.5
        """
        if self._data._dataset_type == 'MultiLabeled_Tiles':
            # Get x, y from validation dataset
            data_loader = self._data.valid_dl
            nbatches = math.ceil(len(self._data.valid_ds)/self._data.batch_size)
            from .._utils.common import get_nbatches
            x_batch, y_batch = get_nbatches(data_loader, nbatches)
            x_batch = torch.cat(x_batch)
            y_batch = torch.cat(y_batch)
            score_thresh = kwargs.get('thresh', 0.5)

            # Get predictions
            predictions = []
            learn_temp = copy.copy(self.learn)
            for i in range(0, x_batch.shape[0], self._data.batch_size):
                batch_preds = learn_temp.pred_batch(batch=(x_batch[i:i+self._data.batch_size], y_batch[i:i+self._data.batch_size]))
                predictions.append(batch_preds)
            predictions = torch.cat(predictions)
            one_hot_preds = (predictions >= score_thresh)

            # Use Scikit-learn multilabel confusion matrix
            from sklearn.metrics import multilabel_confusion_matrix
            y_true = y_batch.to('cpu').numpy()
            y_pred = one_hot_preds.to('cpu').numpy()
            confusion_matrix = multilabel_confusion_matrix(y_true, y_pred)

            # Plot the classwise confusion matrix
            nrows=self._data.c
            plt_size = 4
            fig, axs = plt.subplots(nrows=nrows, figsize=(plt_size, (nrows)*plt_size))
            fig.suptitle('Confusion Matrix', fontsize=16)
            top = 1 - (math.sqrt(16)/math.sqrt(100*nrows*plt_size))
            fig.subplots_adjust(top=top, hspace=0.5)

            for i, (classname, matrix) in enumerate(zip(self._data.classes, confusion_matrix)):
                cm = np.fliplr(np.flipud(matrix))
                axi = axs[i]
                cmap = "Blues"
                axi.imshow(cm, interpolation='nearest', cmap=cmap)
                title = classname
                axi.set_title(title)
                tick_marks = np.arange(2)
                axi.set_xticks(ticks=tick_marks)
                axi.set_xticklabels([classname, 'Rest'])
                axi.set_yticks(ticks=tick_marks)
                axi.set_yticklabels([classname, 'Rest'])
                axi.set_ylabel('Actual')
                axi.set_xlabel('Predicted')
                axi.grid(False)

                import itertools
                thresh = cm.max() / 2.
                for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
                    coeff = f'{cm[i, j]}'
                    axi.text(j, i, coeff, horizontalalignment="center", verticalalignment="center", color="white" if cm[i, j] > thresh else "black")

        # For single label classification
        else:
            self._check_requisites()
            learn_temp = copy.copy(self.learn)
            interp = ClassificationInterpretation.from_learner(learn_temp)
            interp.plot_confusion_matrix()

    def plot_hard_examples(self, num_examples):
        """
        Plots the hard examples with their heatmaps.
        
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        num_examples            Number of hard examples to plot
                                ``prepare_data`` function.
        =====================   ===========================================
        """
        self._check_requisites()
        # handling bug in fastai.
        if num_examples == 1:
            num_examples = 2
        learn_temp = copy.copy(self.learn)
        interp = ClassificationInterpretation.from_learner(learn_temp)
        heatmap = True
        if self._backend == 'tensorflow':
            heatmap = False
        if self._data._dataset_type == 'MultiLabeled_Tiles':
            try:
                interp.plot_multi_top_losses(num_examples, figsize=(5,5))
            except IndexError:
                from IPython.display import clear_output
                clear_output(wait=True)
                print("No mismatches found.")
            return
        fig = interp.plot_top_losses(num_examples, figsize=(15,15), heatmap=heatmap, return_fig=True)
        # fastai way of calculating num nrows and ncols
        cols = math.ceil(math.sqrt(num_examples))
        rows = math.ceil(num_examples/cols)
        axes = fig.axes
        # get number of empty axes from behind.
        num_empty_ax = rows * cols - num_examples
        # delete those from back.
        for k in range(num_empty_ax):
            fig.delaxes(axes[-(k+1)])

    @staticmethod
    def _convert_to_degrees(value, reference):
        d0 = value[0][0]
        d1 = value[0][1]
        d = float(d0) / float(d1)

        m0 = value[1][0]
        m1 = value[1][1]
        m = float(m0) / float(m1)

        s0 = value[2][0]
        s1 = value[2][1]
        s = float(s0) / float(s1)

        degrees = d + (m / 60.0) + (s / 3600.0)

        if reference == "S" or reference == "W":
            degrees = 0 - degrees
        
        return degrees

    def predict_folder_and_create_layer(self, folder, feature_layer_name, gis=None, prediction_field='predict', confidence_field='confidence'):
        """
        Predicts on images present in the given folder and creates a feature layer.
        The images stored in the folder contain GPS information as part of EXIF metadata.
        Works with RGB images only.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        folder                  Required String. Folder containing images to inference on.
        ---------------------   -------------------------------------------
        feature_layer_name      Required String. The name of the feature layer used to publish.   
        ---------------------   -------------------------------------------
        gis                     Optional GIS Object, the GIS on which this tool runs. If not specified,
                                the active GIS is used.      
        ---------------------   -------------------------------------------
        prediction_field        Optional String. The field name to use to add predictions.             
        ---------------------   -------------------------------------------
        confidence_field        Optional String. The field name to use to add confidence.                                                                               
        =====================   ===========================================

        :returns: `FeatureCollection` Object                              
        """        
        return self._create_feature_layer(
            self._extract_images_geo_data(folder),
            _arcgis.env.active_gis if gis is None else gis,
            feature_layer_name,
            prediction_field,
            confidence_field
        )

    def _extract_images_geo_data(self, folder):
        ALLOWED_FILE_FORMATS = ['tif', 'jpg', 'png']

        files = []

        for ext in ALLOWED_FILE_FORMATS:
            files.extend(glob.glob(os.path.join(folder, '*.' + ext)))
        
        images_data = []

        for file in files:
            img = PIL.Image.open(file)
            exif = {
                PIL.ExifTags.TAGS[k]: v
                for k, v in img._getexif().items()
                if k in PIL.ExifTags.TAGS
            }

            images_data.append(
                {
                    'image_path': file,
                    'y': FeatureClassifier._convert_to_degrees(exif['GPSInfo'][2], exif['GPSInfo'][1]),
                    'x': FeatureClassifier._convert_to_degrees(exif['GPSInfo'][4], exif['GPSInfo'][3])
                }
            )
        
        return images_data

    def _create_feature_layer(self, images_data, gis_user, feature_layer_name, prediction_field, confidence_field):
        data = []
        images = {}
        for image_data in images_data:
            prediction = self.predict(image_data['image_path'])
            images[os.path.basename(image_data['image_path'])] = image_data['image_path']
            data.append(
                [   
                    os.path.basename(image_data['image_path']),
                    prediction[0].obj,
                    prediction[2].data.max().tolist(),
                    image_data['x'],
                    image_data['y']
                ]
            )
        
        dataframe = pandas.DataFrame(data, columns=['image_name', prediction_field, confidence_field, 'X', 'Y'])
        spatial_dataframe = dataframe.spatial.from_xy(df=dataframe, sr=4326, x_column='X', y_column='Y')

        feature_collection = gis_user.content.import_data(spatial_dataframe, title=feature_layer_name)

        feature_layer = feature_collection.layers[0]
        feature_layer.manager.add_to_definition({"hasAttachments":True})

        df = feature_layer.query(as_df=True)
        object_field = feature_layer.properties['objectIdField']

        for image_name, image_path in images.items():
            object_id = df.loc[df['image_name'] == image_name, object_field].values[0]  #assuming image_name is unique
            if np.isnan(object_id):
                continue #skipping those values which are not present.
            feature_layer.attachments.add(
                object_id,
                image_path
            )

        return feature_collection

    @staticmethod
    def _update_predictions_layer(feature_layer, features_to_update, output_label_field, confidence_field=None):
        field_template = {
            "name": output_label_field,
            "type": "esriFieldTypeString",
            "alias": output_label_field,
            "sqlType": "sqlTypeOther",
            "length": 256,
            "nullable": True,
            "editable": True,
            "visible": True,
            "domain": None,
            "defaultValue": ''
        }

        confidence_field_template = {
            "name": confidence_field,
            "type": "esriFieldTypeString",
            "alias": confidence_field,
            "sqlType": "sqlTypeOther",
            "length": 256,
            "nullable": True,
            "editable": True,
            "visible": True,
            "domain": None,
            "defaultValue": ''
        }

        feature_layer.manager.add_to_definition({'fields': [field_template]})

        if confidence_field:
            feature_layer.manager.add_to_definition({'fields': [confidence_field_template]})

        try:
            start = 0
            stop = 100
            count = 100

            features_updated = features_to_update[start:stop]
            response = feature_layer.edit_features(updates=features_updated)

            for resp in response.get('updateResults', []):
                if resp.get('success', False):
                    continue
                warnings.warn(f"Something went wrong for data {resp}")

            time.sleep(2)
            while count == len(features_updated):
                start = stop
                stop = stop + 100
                features_updated = features_to_update[start:stop]
                response = feature_layer.edit_features(updates=features_updated)
                for resp in response.get('updateResults', []):
                    if resp.get('success', False):
                        continue
                    warnings.warn(f"Something went wrong for data {resp}")
                time.sleep(2)
        except Exception:
            feature_layer.manager.delete_from_definition({'fields': [field_template]})
            if confidence_field:
                feature_layer.manager.delete_from_definition({'fields': [confidence_field_template]})

            return False

        return True

    def _classify_attachments(
            self,
            feature_layer,
            data_folder,
            feature_attachments_mapping,
            input_label_field,
            output_label_field,
            confidence_field=None,
            predict_function=_prediction_function
    ):

        features = feature_layer.query().features
        features_to_update = []

        for feature in features:
            feature_attachments = (feature_attachments_mapping.get(str(feature.attributes[input_label_field])) or \
                                   feature_attachments_mapping.get(int(feature.attributes[input_label_field])))
            if not feature_attachments:
                continue

            predictions = []
            for attachment in feature_attachments:
                prediction = self.predict(os.path.join(data_folder, attachment))
                predictions.append((prediction[0].obj, prediction[2].data.max().tolist()))

            final_prediction = predict_function(predictions)

            feature.attributes[output_label_field] = final_prediction[0]
            if confidence_field:
                feature.attributes[confidence_field] = final_prediction[1]

            features_to_update.append(feature)

        return features_to_update

    def _classify_labeled_tiles(
            self,
            feature_layer,
            labeled_tiles_directory,
            input_label_field,
            output_label_field,
            confidence_field=None
    ):
        ALLOWED_FILE_FORMATS = ['tif', 'jpg', 'png']
        IMAGES_FOLDER = 'images/'
        LABELS_FOLDER = 'labels/'

        files = []

        for ext in ALLOWED_FILE_FORMATS:
            files.extend(glob.glob(os.path.join(labeled_tiles_directory, IMAGES_FOLDER + '*.' + ext)))

        predictions = {}
        for file in files:
            xml_path = os.path.join(os.path.dirname(os.path.dirname(file)),
                                    os.path.join(LABELS_FOLDER, os.path.basename(file).split('.')[0] + '.xml'))

            if not os.path.exists(xml_path):
                continue

            tree = ElementTree.parse(xml_path)
            root = tree.getroot()

            name_field = root.findall('object/name')
            if len(name_field) != 1:
                continue

            file_prediction = self.predict(file)

            predictions[name_field[0].text] = {
                'prediction': file_prediction[0].obj,
                'score': str(file_prediction[2].data.max().tolist())
            }

        features = feature_layer.query(output_fields=[input_label_field]).features
        features_to_update = []
        for feature in features:
            if predictions.get(str(feature.attributes[input_label_field])):
                feature.attributes[output_label_field] = predictions.get(str(feature.attributes[input_label_field]))[
                    'prediction']
                if confidence_field:
                    feature.attributes[confidence_field] = predictions.get(str(feature.attributes[input_label_field]))[
                        'score']

                features_to_update.append(feature)

        return features_to_update

    def classify_features(
        self,
        feature_layer,
        labeled_tiles_directory,
        input_label_field,
        output_label_field,
        confidence_field=None,
        predict_function=None
    ):

        """
        Classifies the exported images and updates the feature layer with the prediction results in the ``output_label_field``.
        Works with RGB images only.

        ====================================     ====================================================================
        **Argument**                             **Description**
        ------------------------------------     --------------------------------------------------------------------
        feature_layer                            Required. Feature Layer for classification.
        ------------------------------------     --------------------------------------------------------------------
        labeled_tiles_directory                  Required. Folder structure containing images and labels folder. The
                                                 chips should have been generated using the export training data tool in
                                                 the Labeled Tiles format, and the labels should contain the OBJECTIDs
                                                 of the features to be classified.
        ------------------------------------     --------------------------------------------------------------------
        input_label_field                        Required. Value field name which created the labeled tiles. This field
                                                 should contain the OBJECTIDs of the features to be classified. In case of
                                                 attachments this field is not used.
        ------------------------------------     --------------------------------------------------------------------
        output_label_field                       Required. Output column name to be added in the layer which contains predictions.
        ------------------------------------     --------------------------------------------------------------------
        confidence_field                         Optional. Output column name to be added in the layer which contains the confidence score.
        ------------------------------------     --------------------------------------------------------------------
        predict_function                         Optional. Used for calculation of final prediction result when each feature
                                                 has more than one attachment. The ``predict_function`` takes as input a list of tuples.
                                                 Each tuple has first element as the class predicted and second element is the confidence score.
                                                 The function should return the final tuple classifying the feature and its confidence
        ====================================     ====================================================================

        :return:
            Boolean : True/False if operation is successful

        """

        if predict_function is None:
            predict_function = _prediction_function

        if input_label_field and _check_esri_files(Path(labeled_tiles_directory)):
            features_to_update = self._classify_labeled_tiles(
                feature_layer,
                labeled_tiles_directory,
                input_label_field,
                output_label_field,
                confidence_field
            )
        elif os.path.exists(os.path.join(labeled_tiles_directory, 'mapping.txt')):
            json_file = os.path.join(labeled_tiles_directory, 'mapping.txt')
            with open(json_file) as file:
                feature_attachments_mapping = json.load(file)

            features_to_update = self._classify_attachments(
                feature_layer,
                labeled_tiles_directory,
                feature_attachments_mapping,
                feature_layer.properties['objectIdField'],
                output_label_field,
                confidence_field,
                predict_function
            )
        else:
            return False

        return FeatureClassifier._update_predictions_layer(
            feature_layer,
            features_to_update,
            output_label_field,
            confidence_field
        )

    def _categorize_feature_layer(
        self,
        feature_layer,
        raster,
        class_value_field,
        class_name_field,
        confidence_field,
        cell_size,
        coordinate_system,
        predict_function,
        batch_size,
        overwrite
    ):  
        # class values
        class_values = list(self._data.class_mapping.keys())

        # normalization stats
        norm_mean = torch.tensor(imagenet_stats[0])
        norm_std = torch.tensor(imagenet_stats[1])

        # Check and create Fields 
        class_name_field_template = {
            "name": class_name_field.lower(),
            "type": "esriFieldTypeString",
            "alias": class_name_field,
            "sqlType": "sqlTypeOther",
            "length": 256,
            "nullable": True,
            "editable": True,
            "visible": True,
            "domain": None,
            "defaultValue": ''
        }

        class_value_field_template = {
            "name": class_value_field.lower(),
            "type": "esriFieldTypeInteger",
            "alias": class_value_field,
            "sqlType": "sqlTypeOther",
            "nullable": True,
            "editable": True,
            "visible": True,
            "domain": None,
            "defaultValue": -999
        }

        to_delete = []
        to_create = []

        feature_layer_fields = { f['name'].lower():f for f in feature_layer.properties["fields"] }
        oid_field = feature_layer.properties['objectIdField']

        if class_value_field_template['name'] in feature_layer_fields:
            if overwrite:
                to_delete.append(feature_layer_fields[class_value_field_template['name']])
            else:
                e = Exception(f"The specified class_value_field '{class_value_field}' already exists, please specify a different name or set `overwrite=True`")
                raise(e)
        to_create.append(class_value_field_template)

        if class_name_field_template['name'] in feature_layer_fields:
            if overwrite:
                to_delete.append(feature_layer_fields[class_name_field_template['name']])
            else:
                e = Exception(f"The specified class_name_field '{class_name_field}' already exists, please specify a different name or set `overwrite=True`")
                raise(e)
        to_create.append(class_name_field_template)
        
        if confidence_field is not None:
            confidence_field_template = {
                "name": confidence_field.lower(),
                "type": "esriFieldTypeDouble",
                "alias": confidence_field,
                "sqlType": "sqlTypeDouble",
                "nullable": True,
                "editable": True,
                "visible": True,
                "domain": None,
                "defaultValue": -999
            }
            if confidence_field_template['name'] in feature_layer_fields:
                if overwrite:
                    to_delete.append(feature_layer_fields[confidence_field_template['name']])
                else:
                    e = Exception(f"The specified confidence_field '{confidence_field}' already exists, please specify a different name or set `overwrite=True`")
                    raise(e)
            to_create.append(confidence_field_template)

        feature_layer.manager.delete_from_definition({'fields': to_delete})
        feature_layer.manager.add_to_definition({'fields': to_create})

        # Get features for updation
        fields_to_update = [oid_field, class_value_field, class_name_field]
        if confidence_field is not None:
            fields_to_update.append(confidence_field)
        feature_layer_features = feature_layer.query(out_fields=",".join(fields_to_update), return_geometry=False).features
        update_store = {}

        if raster is not None:
            if not HAS_ARCPY:
                raise Exception("This function requires arcpy.")

            #Arcpy Environment to export data
            arcpy.env.cellSize = cell_size
            if coordinate_system is not None:
                arcpy.env.outputCoordinateSystem = coordinate_system
                arcpy.env.cartographicCoordinateSystem = coordinate_system

            feature_layer_url = feature_layer.url

            if feature_layer._token is not None:
                feature_layer_url = feature_layer_url + f"?token={feature_layer._token}"
            
            # Create Temporary ID field
            tempid_field = _tempid_field = 'f_fcuid'
            i = 1
            while tempid_field in feature_layer_fields:
                tempid_field = _tempid_field + str(i)
                i+=1
            #arcpy.AddField_management(feature_layer_url, tempid_field, "LONG")
            tempid_field_template = {
                "name": tempid_field,
                "type": "esriFieldTypeInteger",
                "alias": tempid_field,
                "sqlType": "sqlTypeOther",
                "nullable": True,
                "editable": True,
                "visible": True,
                "domain": None,
                "defaultValue": -999
            }
            feature_layer.manager.add_to_definition({'fields': [tempid_field_template]})
            #arcpy.CalculateField_management(feature_layer_url, tempid_field, f"{oid_field}", "SQL")
            feature_layer.calculate(where='1=1', calc_expression={"field": tempid_field, "sqlExpression": f"{oid_field}"})

            temp_folder = arcpy.env.scratchFolder
            temp_datafldr = os.path.join(temp_folder, 'categorize_features_'+str(int(time.time())))
            result = arcpy.ia.ExportTrainingDataForDeepLearning(
                in_raster=raster,
                out_folder=temp_datafldr,
                in_class_data=feature_layer_url,
                image_chip_format="TIFF",
                tile_size_x=self._data.chip_size,
                tile_size_y=self._data.chip_size,
                stride_x=0,
                stride_y=0,
                output_nofeature_tiles="ALL_TILES",
                metadata_format="Labeled_Tiles",
                start_index=0,
                class_value_field=tempid_field,
                buffer_radius=0,
                in_mask_polygons=None,
                rotation_angle=0
            )
            # cleanup
            #arcpy.DeleteField_management(feature_layer_url, [ tempid_field ])
            feature_layer.manager.delete_from_definition({'fields': [tempid_field_template]})

            image_list = ImageList.from_folder(os.path.join(temp_datafldr, 'images'))
            def get_id(imagepath):
                with open(os.path.join(temp_datafldr, 'labels', os.path.basename(imagepath)[:-3]+'xml')) as f:
                    return(int(f.read().split('<name>')[1].split('<')[0]))

            for i in range(0, len(image_list), batch_size):
                # Get Temporary Ids
                tempids = [ get_id(f) for f in image_list.items[i:i+batch_size] ]
                
                # Get Image batch
                image_batch = torch.stack([ im.data for im in image_list[i:i+batch_size] ])
                image_batch = normalize(image_batch, mean=norm_mean, std=norm_std)
                
                # Get Predications
                predicted_classes, predictions_conf = self._predict_batch(image_batch)
                
                # push prediction to store
                for ui, oid in enumerate(tempids):
                    classvalue = class_values[predicted_classes[ui]]
                    update_store[oid] = {
                        oid_field: oid,
                        class_value_field: classvalue,
                        class_name_field:  self._data.class_mapping[classvalue]
                    } 
                    if confidence_field is not None:
                        update_store[oid][confidence_field] = predictions_conf[ui]

            # Cleanup
            arcpy.Delete_management(temp_datafldr)
            shutil.rmtree(temp_datafldr, ignore_errors=True)

        else:
            out_folder = tempfile.TemporaryDirectory().name
            os.mkdir(out_folder)
            feature_layer.export_attachments(out_folder)
            with open(os.path.join(out_folder, 'mapping.txt')) as file:
                feature_attachments_mapping = json.load(file)
                images_store = []
                for oid in feature_attachments_mapping:
                    for im in feature_attachments_mapping[oid]:
                        images_store.append({
                            'oid': oid,
                            'im': os.path.join(out_folder,im)
                        })
            update_store_scratch = {}
            for i in range(0, len(images_store), batch_size):
                rel_objectids = []
                image_batch = []
                for r in images_store[i:i+batch_size]:
                    im = open_image(r['im']) # Read Bytes
                    im = im.resize(self._data.chip_size) # Resize
                    image_batch.append(im.data) # Convert to tensor
                    rel_objectids.append(int(r['oid']))
                image_batch = torch.stack(image_batch)
                image_batch = normalize(image_batch, mean=norm_mean, std=norm_std)
                # Get Predictions and save to scratch
                predicted_classes, predictions_conf = self._predict_batch(image_batch)
                for ai, oid in enumerate(rel_objectids):
                    if update_store_scratch.get(oid) is None:
                        update_store_scratch[oid] = []
                    update_store_scratch[oid].append([predicted_classes[ai], predictions_conf[ai]])
            # Prepare final updated features
            for oid in update_store_scratch:
                max_prediction_class, max_prediction_value = predict_function(update_store_scratch[oid])
                if max_prediction_class is not None:
                    classvalue = class_values[max_prediction_class]
                    classname = self._data.class_mapping[classvalue]
                else:
                    classvalue = None
                    classname = None                
                update_store[oid] = {
                    oid_field: oid,
                    class_value_field: classvalue,
                    class_name_field:  classname
                } 
                if confidence_field is not None:
                    update_store[oid][confidence_field] = max_prediction_value

        # Update Features
        features_to_update = []
        for feat in feature_layer_features:
            if update_store.get(feat.attributes[oid_field]) is not None:
                updated_attributes = update_store[feat.attributes[oid_field]]
                for f in fields_to_update:
                    feat.attributes[f] = updated_attributes[f]
                features_to_update.append(feat)
        step = 100
        for si in range(0, len(features_to_update), step):
            feature_batch = features_to_update[si:si+step]
            response = feature_layer.edit_features(updates=feature_batch)
            for resp in response.get('updateResults', []):
                if resp.get('success', False):
                    continue
                warnings.warn(f"Something went wrong for data {resp}")
            time.sleep(2)
        return True


    def _categorize_feature_class(
        self,
        feature_class,
        raster,
        class_value_field,
        class_name_field,
        confidence_field,
        cell_size,
        coordinate_system,
        predict_function,
        batch_size,
        overwrite
    ):
        
        # class values
        class_values = list(self._data.class_mapping.keys())

        if not HAS_ARCPY:
            raise Exception("This function requires arcpy to access feature class.")
        arcpy.env.overwriteOutput = overwrite

        if batch_size is None:
            batch_size  = self._data.batch_size

        if predict_function is None:
            predict_function = _prediction_function
        
        norm_mean = torch.tensor(imagenet_stats[0])
        norm_std = torch.tensor(imagenet_stats[1])
        
        fcdesc = arcpy.Describe(feature_class)
        oid_field = fcdesc.OIDFieldName
        if not (fcdesc.dataType == 'FeatureClass' and fcdesc.shapeType == 'Polygon'):
            e = Exception(f"The specified FeatureClass at '{feature_class}' is not valid, it should be Polygon FeatureClass")
            raise(e)
        fields = arcpy.ListFields(feature_class)
        field_names = [f.name for f in fields]
        if class_value_field in field_names:
            if not overwrite:
                e = Exception(f"The specified class_value_field '{class_value_field}' already exists in the target FeatureClass, please specify a different name or set `overwrite=True`")
                raise(e)
        arcpy.DeleteField_management(feature_class, 
                                [ class_value_field ])
        arcpy.AddField_management(feature_class, class_value_field, "LONG")
            
        if class_name_field in field_names:
            if not overwrite:
                e = Exception(f"The specified class_name_field '{class_name_field}' already exists in the target FeatureClass, please specify a different name or set `overwrite=True`")
                raise(e)
        arcpy.DeleteField_management(feature_class, 
                                [ class_name_field ])
        arcpy.AddField_management(feature_class, class_name_field, "TEXT")

        if confidence_field is not None:
            if confidence_field in field_names:
                if not overwrite:
                    e = Exception(f"The specified confidence_field '{confidence_field}' already exists in the target FeatureClass, please specify a different name or set `overwrite=True`")
                    raise(e)
            arcpy.DeleteField_management(feature_class, 
                                    [ confidence_field ])
            arcpy.AddField_management(feature_class, confidence_field, "DOUBLE")

        if raster is not None:
            #Arcpy Environment to export data
            arcpy.env.cellSize = cell_size
            if coordinate_system is not None:
                arcpy.env.outputCoordinateSystem = coordinate_system
                arcpy.env.cartographicCoordinateSystem = coordinate_system

            tempid_field = _tempid_field = 'f_fcuid'
            i = 1
            while tempid_field in field_names:
                tempid_field = _tempid_field + str(i)
                i+=1
            arcpy.AddField_management(feature_class, tempid_field, "LONG")
            arcpy.CalculateField_management(feature_class, tempid_field, f"!{oid_field}!")

            temp_folder = arcpy.env.scratchFolder
            temp_datafldr = os.path.join(temp_folder, 'categorize_features_'+str(int(time.time())))
            result = arcpy.ia.ExportTrainingDataForDeepLearning(
                in_raster=raster,
                out_folder=temp_datafldr,
                in_class_data=feature_class,
                image_chip_format="TIFF",
                tile_size_x=self._data.chip_size,
                tile_size_y=self._data.chip_size,
                stride_x=0,
                stride_y=0,
                output_nofeature_tiles="ALL_TILES",
                metadata_format="Labeled_Tiles",
                start_index=0,
                class_value_field=tempid_field,
                buffer_radius=0,
                in_mask_polygons=None,
                rotation_angle=0
            )
            # cleanup
            arcpy.DeleteField_management(feature_class, [ tempid_field ])
            image_list = ImageList.from_folder(os.path.join(temp_datafldr, 'images'))
            def get_id(imagepath):
                with open(os.path.join(temp_datafldr, 'labels', os.path.basename(imagepath)[:-3]+'xml')) as f:
                    return(int(f.read().split('<name>')[1].split('<')[0]))

            for i in range(0, len(image_list), batch_size):
                # Get Temporary Ids
                tempids =[ get_id(f) for f in image_list.items[i:i+batch_size] ]
                
                # Get Image batch
                image_batch = torch.stack([ im.data for im in image_list[i:i+batch_size] ])
                image_batch = normalize(image_batch, mean=norm_mean, std=norm_std)
                
                # Get Predications
                predicted_classes, predictions_conf = self._predict_batch(image_batch)
                
                # Update Feature Class
                where_clause = f"{oid_field} IN ({','.join(str(e) for e in tempids)})"
                update_cursor = arcpy.UpdateCursor(
                    feature_class,
                    where_clause=where_clause,
                    sort_fields=f"{oid_field} A"
                )
                for row in update_cursor:
                    row_tempid = row.getValue(oid_field)
                    ui = tempids.index(row_tempid)
                    classvalue = class_values[predicted_classes[ui]]
                    row.setValue(class_value_field, classvalue)
                    row.setValue(class_name_field, self._data.class_mapping[classvalue])
                    if confidence_field is not None:
                        row.setValue(confidence_field, predictions_conf[ui])
                    update_cursor.updateRow(row)

                # Remove Locks
                del row
                del update_cursor

            # Cleanup
            arcpy.Delete_management(temp_datafldr)
            shutil.rmtree(temp_datafldr, ignore_errors=True)

        else:
            feature_class_attach = feature_class+'__ATTACH'
            nrows = arcpy.GetCount_management(feature_class_attach)[0]
            store={}
            for i in range(0, int(nrows), batch_size):
                attachment_ids = []
                rel_objectids = []
                image_batch = []
                
                # Get Image Batch
                with arcpy.da.SearchCursor(feature_class_attach, [ 'ATTACHMENTID', 'REL_OBJECTID', 'DATA' ]) as search_cursor:
                    for c, item in enumerate(search_cursor):
                        if c >= i and c < i+batch_size :
                            attachment_ids.append(item[0])
                            rel_objectids.append(item[1])
                            attachment = item[-1]
                            im = open_image(io.BytesIO(attachment.tobytes())) # Read Bytes
                            im = im.resize(self._data.chip_size) # Resize
                            image_batch.append(im.data) # Convert to tensor
                            del item
                            del attachment
                            #del im
                image_batch = torch.stack(image_batch)
                image_batch = normalize(image_batch, mean=norm_mean, std=norm_std)
                
                # Get Predictions and save to store
                predicted_classes, predictions_conf = self._predict_batch(image_batch)
                for ai in range(len(attachment_ids)):
                    if store.get(rel_objectids[ai]) is None:
                        store[rel_objectids[ai]] = []
                    store[rel_objectids[ai]].append([predicted_classes[ai], predictions_conf[ai]])
                
            # Update Feature Class
            update_cursor = arcpy.UpdateCursor(feature_class)
            for row in update_cursor:
                row_oid = row.getValue(oid_field)
                max_prediction_class, max_prediction_value = predict_function(store[row_oid])
                if max_prediction_class is not None:
                    classvalue = class_values[max_prediction_class]
                    classname = self._data.class_mapping[classvalue]
                else:
                    classvalue = None
                    classname = None
                row.setValue(class_value_field, classvalue)
                row.setValue(class_name_field, classname)
                if confidence_field is not None:
                    row.setValue(confidence_field, max_prediction_value)
                update_cursor.updateRow(row)

            # Remove Locks
            del row
            del update_cursor
        return True

    @deprecated(deprecated_in="1.7.1", details="Please use arcgis.learn.classify_objects() instead")
    def categorize_features(
        self,
        feature_layer,
        raster=None,
        class_value_field='class_val',
        class_name_field='prediction',
        confidence_field="confidence",
        cell_size=1,
        coordinate_system=None,
        predict_function=None,
        batch_size=64,
        overwrite=False 
    ):
        """
        Categorizes each feature by classifying its attachments or an image of its geographical area (using the provided Imagery Layer)
        and updates the feature layer with the prediction results in the ``output_label_field``.
        Deprecated, please use arcgis.learn.classify_objects() instead.

        ====================================     ====================================================================
        **Argument**                             **Description**
        ------------------------------------     --------------------------------------------------------------------
        feature_layer                            Required. Public Feature Layer or path of local feature class for classification with read, write, edit permissions.
        ------------------------------------     --------------------------------------------------------------------
        raster                                   Optional. Imagery layer or path of local raster to be used for exporting image chips. (Requires arcpy)
        ------------------------------------     --------------------------------------------------------------------
        class_value_field                        Required string. Output field to be added in the layer, containing class value of predictions.
        ------------------------------------     --------------------------------------------------------------------
        class_name_field                         Required string. Output field to be added in the layer, containing class name of predictions.
        ------------------------------------     --------------------------------------------------------------------
        confidence_field                         Optional string. Output column name to be added in the layer which contains the confidence score.
        ------------------------------------     --------------------------------------------------------------------
        cell_size                                Optional float. Cell size to be used for exporting the image chips.
        ------------------------------------     --------------------------------------------------------------------
        coordinate_system                        Optional. Cartographic Coordinate System to be used for exporting the image chips.
        ------------------------------------     --------------------------------------------------------------------
        predict_function                         Optional list of tuples. Used for calculation of final prediction result when each feature
                                                 has more than one attachment. The ``predict_function`` takes as input a list of tuples.
                                                 Each tuple has first element as the class predicted and second element is the confidence score.
                                                 The function should return the final tuple classifying the feature and its confidence.
        ------------------------------------     --------------------------------------------------------------------
        batch_size                               Optional integer. The no of images or tiles to process in a single go. 
                                                 
                                                 The default value is 64.
        ------------------------------------     --------------------------------------------------------------------
        overwrite                                Optional boolean. If set to True the output fields will be overwritten by new values. 

                                                 The default value is False.
        ====================================     ====================================================================

        :return:
            Boolean : True if operation is successful, False otherwise

        """        

        import arcgis
        from arcgis.features import FeatureLayer
        from arcgis.raster import ImageryLayer
        from arcgis.gis import Item
        
        class_value_field = class_value_field.lower()
        class_name_field = class_name_field.lower()
        confidence_field = confidence_field.lower()

        if predict_function is None:
            predict_function = _prediction_function

        if isinstance(raster, str):
            if 'http' in raster:
                raster = ImageryLayer(raster, gis=arcgis.env.active_gis)
        if isinstance(raster, ImageryLayer):
            raster_url = raster.url
            if raster._token is not None:
                raster_url = raster_url + f"?token={raster._token}" 
            raster = raster_url
        elif isinstance(raster, Item):  # handles Mapserver
            raster_url = raster.url
            if raster.layers[0]._token is not None:
                raster_url = raster_url + f"?token={raster.layers[0]._token}" 
            raster = raster_url

        if isinstance(feature_layer, str):
            if 'http' in feature_layer:
                feature_layer = FeatureLayer(feature_layer, gis=arcgis.env.active_gis)
            else:
                return self._categorize_feature_class(
                    feature_class=feature_layer,
                    raster=raster,
                    class_value_field=class_value_field,
                    class_name_field=class_name_field,
                    confidence_field=confidence_field,
                    cell_size=cell_size,
                    coordinate_system=coordinate_system,
                    predict_function=predict_function,
                    batch_size=batch_size,
                    overwrite=overwrite
                )

        if isinstance(feature_layer, FeatureLayer):
            return self._categorize_feature_layer(
                feature_layer=feature_layer,
                raster=raster,
                class_value_field=class_value_field,
                class_name_field=class_name_field,
                confidence_field=confidence_field,
                cell_size=cell_size,
                coordinate_system=coordinate_system,
                predict_function=predict_function,
                batch_size=batch_size,
                overwrite=overwrite
            )
        else:
            e = Exception("Could not understand layer type")
            raise(e)
     
    ## Tensorflow specific functions start ##
    def _intialize_tensorflow(self, data, backbone, drop, pretrained_path, kwargs):
        self._check_tf()

        from .._utils.fastai_tf_fit import TfLearner
        import tensorflow as tf
        from tensorflow.keras.losses import CategoricalCrossentropy
        from tensorflow.keras.models import Model
        from tensorflow.keras import applications
        from tensorflow.keras.optimizers import Adam
        from fastai.basics import defaults
        from .._utils.image_classification import TF_IC_get_head_output

        if data._is_multispectral:
            raise Exception('Multispectral data is not supported with backend="tensorflow"')

        # Pyramid Scheme in head
        self._fpn = kwargs.get('fpn', True)

        # prepare color array
        alpha = 0.7
        color_mapping = getattr(data, 'color_mapping', None)
        if color_mapping is None:
            color_array = torch.tensor([[1., 1., 1.]]).float()
        else:
            color_array = torch.tensor( list(color_mapping.values()) ).float() / 255
        alpha_tensor = torch.tensor( [alpha]*len(color_array) ).view(-1, 1).float()
        color_array = torch.cat( [ color_array, alpha_tensor ], dim=-1)
        background_color = torch.tensor( [[0, 0, 0, 0]] ).float()
        data._multispectral_color_array = torch.cat( [background_color, color_array] )

        self.ssd_version = 1#ssd_version
        if backbone is None:
            backbone = 'ResNet50'
        
        if type(backbone) == str:
            backbone = getattr(applications, backbone)
            
        self._backbone = backbone 
        
        x, y = next(iter(data.train_dl))
        if tf.keras.backend.image_data_format() == 'channels_last':
            in_shape = [x.shape[-1], x.shape[-1], 3]
        else:
            in_shape = [3, x.shape[-1],x.shape[-1]]

        self._backbone_initalized = self._backbone(
            input_shape=in_shape, 
            include_top=False, 
            weights='imagenet'
        )
        self._backbone_initalized.trainable = False

        self._device = torch.device('cpu')
        self._data = data

        self._loss_function_tf_ = CategoricalCrossentropy(from_logits=True, reduction='sum')
        self._loss_function_tf_noreduction = CategoricalCrossentropy(from_logits=True, reduction=tf.keras.losses.Reduction.NONE)
        
        output_layer = TF_IC_get_head_output(self)

        model = Model(inputs=self._backbone_initalized.input, outputs=output_layer)

        self.learn = TfLearner(
            data, 
            model,
            opt_func=Adam,
            loss_func=self._loss_function_tf,
            true_wd=True, 
            bn_wd=True, 
            wd=defaults.wd, 
            train_bn=True
        )
        
        self.learn.unfreeze()
        self.learn.freeze_to(len(self._backbone_initalized.layers))

        self.show_results = self._show_results_multispectral

        self._code = feature_classifier_prf

    def _loss_function_tf(self, target, predictions, reduction=True):
        import tensorflow as tf
        if target.ndim == 2:
            target_masks = target
        else:
            target_masks = tf.gather(tf.eye(self._data.c), target)
        if reduction:
            return self._loss_function_tf_(target_masks, predictions)
        else:
            return self._loss_function_tf_noreduction(target_masks, predictions)
            
       
if HAS_FASTAI:
    class OverSamplingCallback(LearnerCallback):
        """
        The OverSamplingCallback support handles unbalanced dataset (dataset with rare classes). It is used to oversample data during training.
        """
        def __init__(self,learn:Learner,weights:torch.Tensor=None):
            super().__init__(learn)
            self.weights = weights

        def on_train_begin(self, **kwargs):
            ds,dl = self.data.train_ds,self.data.train_dl
            self.labels = ds.y.items.astype(int)
            assert np.issubdtype(self.labels.dtype, np.integer), "Can only oversample integer values"
            _,self.label_counts = np.unique(self.labels,return_counts=True)
            if self.weights is None: self.weights = torch.DoubleTensor((1/self.label_counts)[self.labels])
            self.total_len_oversample = int(self.data.c*np.max(self.label_counts))
            sampler = WeightedRandomSampler(self.weights, self.total_len_oversample)
            self.data.train_dl = dl.new(shuffle=False, sampler=sampler)



import json
from arcgis.gis import Layer, _GISResource, Item, GIS
###########################################################################
class Object3DLayer(Layer):
    """
    Represents a Web scene Point Cloud layer. Point Cloud layer are cached web layers that are optimized for displaying a large
    amount of 2D and 3D features. You can use scene layers to represent 3D points, point clouds, 3D objects and
    integrated mesh layers.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.Point3DLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """
    def __init__(self, url, gis=None):
        """
        Constructs a SceneLayer given a web scene layer URL
        """
        super(Object3DLayer, self).__init__(url, gis)

    @property
    def _lyr_dict(self):
        url = self.url

        lyr_dict =  { 'type' : "SceneLayer",
                      'url' : url }
        if self._token is not None:
            lyr_dict['serviceToken'] = self._token

        if self.filter is not None:
            lyr_dict['filter'] = self.filter
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict
    #----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += '?token=' + self._token

        lyr_dict = {'type': "SceneLayer", 'url': url}

        if self.filter is not None:
            lyr_dict['options'] = json.dumps({ "definition_expression": self.filter })
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict

###########################################################################
class IntegratedMeshLayer(Layer):
    """
    Represents a Web scene Point Cloud layer. Point Cloud layer are cached web layers that are optimized for displaying a large
    amount of 2D and 3D features. You can use scene layers to represent 3D points, point clouds, 3D objects and
    integrated mesh layers.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.Point3DLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """
    def __init__(self, url, gis=None):
        """
        Constructs a SceneLayer given a web scene layer URL
        """
        super(IntegratedMeshLayer, self).__init__(url, gis)

    @property
    def _lyr_dict(self):
        url = self.url

        lyr_dict =  { 'type' : "IntegratedMeshLayer",
                      'url' : url }
        if self._token is not None:
            lyr_dict['serviceToken'] = self._token

        if self.filter is not None:
            lyr_dict['filter'] = self.filter
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict
    #----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += '?token=' + self._token

        lyr_dict = {'type': "IntegratedMeshLayer", 'url': url}

        if self.filter is not None:
            lyr_dict['options'] = json.dumps({ "definition_expression": self.filter })
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict

###########################################################################
class Point3DLayer(Layer):
    """
    Represents a Web scene Point Cloud layer. Point Cloud layer are cached web layers that are optimized for displaying a large
    amount of 2D and 3D features. You can use scene layers to represent 3D points, point clouds, 3D objects and
    integrated mesh layers.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.Point3DLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """
    def __init__(self, url, gis=None):
        """
        Constructs a SceneLayer given a web scene layer URL
        """
        super(Point3DLayer, self).__init__(url, gis)
    #----------------------------------------------------------------------
    @property
    def _lyr_dict(self):
        url = self.url

        lyr_dict =  { 'type' : "SceneLayer",
                      'url' : url }
        if self._token is not None:
            lyr_dict['serviceToken'] = self._token

        if self.filter is not None:
            lyr_dict['filter'] = self.filter
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict
    #----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += '?token=' + self._token

        lyr_dict = {'type': "SceneLayer", 'url': url}

        if self.filter is not None:
            lyr_dict['options'] = json.dumps({ "definition_expression": self.filter })
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict

###########################################################################
class PointCloudLayer(Layer):
    """
    Represents a Web scene Point Cloud layer. Point Cloud layer are cached web layers that are optimized for displaying a large
    amount of 2D and 3D features. You can use scene layers to represent 3D points, point clouds, 3D objects and
    integrated mesh layers.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.PointCloudLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """
    def __init__(self, url, gis=None):
        """
        Constructs a SceneLayer given a web scene layer URL
        """
        super(PointCloudLayer, self).__init__(url, gis)

    @property
    def _lyr_dict(self):
        url = self.url

        lyr_dict =  { 'type' : 'PointCloudLayer',
                      'url' : url }
        if self._token is not None:
            lyr_dict['serviceToken'] = self._token

        if self.filter is not None:
            lyr_dict['filter'] = self.filter
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict
    #----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += '?token=' + self._token

        lyr_dict = {'type': 'PointCloudLayer', 'url': url}

        if self.filter is not None:
            lyr_dict['options'] = json.dumps({ "definition_expression": self.filter })
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict
###########################################################################
class BuildingLayer(Layer):
    """
    Represents a Web scene layer. Web scene layers are cached web layers that are optimized for displaying a large
    amount of 2D and 3D features. You can use scene layers to represent 3D points, point clouds, 3D objects and
    integrated mesh layers.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.BuildingLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """
    def __init__(self, url, gis=None):
        """
        Constructs a SceneLayer given a web scene layer URL
        """
        super(BuildingLayer, self).__init__(url, gis)

    @property
    def _lyr_dict(self):
        url = self.url

        lyr_dict =  { 'type' : 'BuildingSceneLayer',
                      'url' : url }
        if self._token is not None:
            lyr_dict['serviceToken'] = self._token

        if self.filter is not None:
            lyr_dict['filter'] = self.filter
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict
    #----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += '?token=' + self._token

        lyr_dict = {'type': 'BuildingSceneLayer', 'url': url}

        if self.filter is not None:
            lyr_dict['options'] = json.dumps({ "definition_expression": self.filter })
        if self._time_filter is not None:
            lyr_dict['time'] = self._time_filter
        return lyr_dict
###########################################################################
class _SceneLayerFactory(type):
    """
    Factory that generates the Scene Layers

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.PointCloudLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """
    def __call__(cls,
                 url,
                 gis=None):
        lyr = Layer(url=url, gis=gis)
        props = lyr.properties
        if 'sublayers' in props:
            return BuildingLayer(url=url, gis=gis)
        elif 'layerType' in props:
            lt = props.layerType
        else:
            lt = props.layers[0].layerType
        if str(lt).lower() == "pointcloud":
            return PointCloudLayer(url=url, gis=gis)
        elif str(lt).lower() == "point":
            return Point3DLayer(url=url, gis=gis)
        elif str(lt).lower() == "3dobject":
            return Object3DLayer(url=url, gis=gis)
        elif str(lt).lower() == "building":
            return BuildingLayer(url=url, gis=gis)
        elif str(lt).lower() == "IntegratedMesh".lower():
            return IntegratedMeshLayer(url=url, gis=gis)
        return lyr
###########################################################################
class SceneLayer(Layer, metaclass=_SceneLayerFactory):
    """
    Represents a Web scene layer. Web scene layers are cached web layers that are optimized for displaying a large
    amount of 2D and 3D features. You can use scene layers to represent 3D points, point clouds, 3D objects and
    integrated mesh layers.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.PointCloudLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """
    def __init__(self, url, gis=None):
        """
        Constructs a SceneLayer given a web scene layer URL
        """
        super(SceneLayer, self).__init__(url, gis)



""" Defines store functions for working with Projects.
"""
import arcgis
from arcgis.gis import Item
from arcgis.apps import workforce
from arcgis.apps.workforce._store._definitions import *
from arcgis.apps.workforce.exceptions import WorkforceError
from warnings import warn
import concurrent.futures
import os


def get_project(project_id, gis):
    """ Loads and returns a workforce project.
        :param gis: An authenticated arcigs.gis.GIS object.
        :param project_id: The project's id. Version 1 - id of the project item. Version 2 - id of the feature service
        :returns: workforce.Project
    """
    item = Item(gis, project_id)
    return workforce.Project(item)


def create_project(title, summary=None, major_version=None, gis=None):
    """
        Creates a new Workforce Project

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        title                  :class:`String`.
                               The title of the Project to create (must be unique to the organization)
        ------------------     --------------------------------------------------------------------
        summary                :class:`String`.
                               The summary of the Project
        ------------------     --------------------------------------------------------------------
        major_version          Optional :class:`Int`
                               The version of the Project to create. 1 represents the original
                               Workforce Project which does not support offline. 2 represents the newer
                               Workforce Project which supports offline among other things. Defaults
                               to 2 in GIS 8.2 and higher
        ------------------     --------------------------------------------------------------------
        gis                    Optional :class:`~arcgis.gis.GIS`.
                               The authenticated GIS to use.
                               Defaults to the active GIS if None is provided.
        ==================     ====================================================================

        Returns a :class:`arcgis.apps.workforce.Project`
    """

    if gis is None:
        gis = arcgis.env.active_gis
    
    if major_version is None:
        if gis.version < [8, 2]:
            major_version = 1
        else:
            major_version = 2

    if major_version == 1:
        return _v1_create_project(gis, summary, title)

    elif major_version == 2:
        if gis.version < [7, 1]:
            raise ValueError("Offline Workforce Projects not supported at GIS lower than 7.1")
        return _v2_create_project(gis, summary, title)

    else:
        raise WorkforceError("Invalid Project Version Specified")


def _get_basemap(gis):
    """
    Helper method to get the basemap to use. Tries to use the org defined basemap.
    :param gis:
    :return:
    """
    default_raster_basemap = {
        "baseMapLayers": [
            {
                "id": 'World_Topo_Map',
                "layerType": 'ArcGISTiledMapServiceLayer',
                "url": 'https://services.arcgisonline.com/ArcGIS/rest/services/World_Topo_Map/MapServer',
                "visibility": True,
                "opacity": 1,
                "title": 'World_Topo_Map'
            }
        ],
        "title": 'Topographic'
    }
    try:
        if gis.properties['useVectorBasemaps']:
            bm_group = gis.groups.search(gis.properties['vectorBasemapGalleryGroupQuery'], outside_org=True)[0]
            query = 'group:{} AND name: "0220_Navigation_Title"'.format(bm_group.id)
            items = gis.content.search(query, outside_org=True)
            if items:
                return items[0].get_data()['baseMap']
        else:
            return default_raster_basemap
    except:
        return default_raster_basemap


def _create_application_properties():
    """
    Helper method to build the application properties
    :return:
    """
    return {
        "viewing": {
            "search": {
                "enabled": True,
                "disablePlaceFinder": True,
                "layers": [
                    {
                        "id": "Workers_0",
                        "field": {
                            "name": 'name',
                            "exactMatch": True,
                            "type": 'esriFieldTypeString'
                        }
                    }
                ]
            }
        }
    }


def _build_operational_layers(item, popup_def=None, visibility=True, layer_index=0, capabilities=None):
    """
    Helper method to build operational layers
    :param item: The item containing the layer (at index 0)
    :param popup_def: The popup definition of the layer
    :param visibility: Is this layer visible in the map
    :return: The operational layer dictionary
    """
    layer = item.layers[layer_index]
    op_layer = {
        "layerType": 'ArcGISFeatureLayer',
        "opacity": 1,
        "itemId": item.id,
        "id": "{}_0".format(layer.properties["name"]),
        "title": layer.properties["name"],
        "url": layer.url,
        "visibility": visibility
    }
    if capabilities:
        op_layer["capabilities"] = capabilities
    if popup_def:
        op_layer["popupInfo"] = popup_def
    return op_layer


def _build_table(item, table_index):
    """
    Helper method to build table
    :param item: The item containing the layer (at index 0)
    :param table_index: The index for the table in the item
    """
    table = item.tables[table_index]
    op_table = {
        "capabilities": "Create,Delete,Query,Update,Editing,Sync",
        "url": table.url,
        "id": "{}_0".format(table.properties["name"]),
        "title": table.properties["name"],
        "itemId": item.id,
    }
    return op_table


def _v2_create_project(gis, summary, title):
    """Creates project following version 2 Workforce schema"""
    for f in gis.users.me.folders:
        if f['title'].lower() == title.lower():
            raise WorkforceError("A folder named '{}' already exists.".format(title))

    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        future_folder = executor.submit(gis.content.create_folder, title)
        future_group = executor.submit(gis.groups.create,
                                       title,
                                       'workforce',
                                       access='private',
                                       is_invitation_only=True,
                                       is_view_only=True,
                                       sort_field='title',
                                       sort_order='asc'
                                       )
    future_folder.result()
    group = future_group.result()
    group.protected = True
    folder_name = title
    group_id = group.id
    workforce_service_name = "workforce_{}".format(group_id)
    workforce_service_item = _v2_create_service_with_layers(gis, folder_name, workforce_service_name,
                                                            assignment_layer_definition_v2,
                                                            worker_layer_definition_v2,
                                                            dispatcher_table_definition_v2,
                                                            assignment_type_table_definition_v2,
                                                            app_integration_table_definition_v2,
                                                            title)

    # create webmaps
    my_path = os.path.abspath(os.path.dirname(__file__))
    thumbnail = os.path.join(my_path, '/'.join(('resources', 'default-project-thumbnail.png')))
    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        workers_webmap_future = executor.submit(_v2_create_worker_webmap,
                                                gis,
                                                folder_name,
                                                workforce_service_item,
                                                assignment_layer_popup_definition_v2,
                                                worker_layer_popup_definition_v2,
                                                title, summary, thumbnail)
        dispatchers_webmap_future = executor.submit(_v2_create_dispatcher_webmap,
                                                    gis,
                                                    folder_name,
                                                    workforce_service_item,
                                                    assignment_layer_popup_definition_v2,
                                                    worker_layer_popup_definition_v2,
                                                    title, summary, thumbnail)
    workers_webmap = workers_webmap_future.result()
    dispatchers_webmap = dispatchers_webmap_future.result()

    project_items = [
        workforce_service_item,
        workers_webmap,
        dispatchers_webmap
    ]
    # share and protect items
    for i in project_items:
        i.share(groups=[group])
        i.protect()

    # set fs item properties / thumbnail
    workforce_service_item.update(thumbnail=thumbnail,
                                  item_properties={
                                    "snippet": summary,
                                    "properties": {
                                        "workforceProjectGroupId": group_id,
                                        "workforceProjectVersion": "2.0.0",
                                        "workforceDispatcherMapId": dispatchers_webmap.id,
                                        "workforceWorkerMapId": workers_webmap.id
                                    }
    })

    # manually add the owner as the first dispatcher
    user_id_field_name = "userId"
    for field in workforce_service_item.tables[0].properties.fields:
        if field['name'].lower() == "userid":
            user_id_field_name = field['name']
            break

    workforce_service_item.tables[0].edit_features(adds=[arcgis.features.Feature(
        attributes={
            "name": gis.users.me.fullName,
            user_id_field_name: gis.users.me.username
        }
    )])

    # add relationships
    try:
        workers_webmap.add_relationship(workforce_service_item, 'WorkforceMap2FeatureService')
        dispatchers_webmap.add_relationship(workforce_service_item, 'WorkforceMap2FeatureService')
    except Exception:
        warn("Relationship not added. This version of ArcGIS may not support the WorkforceMap2FeatureService relationship")

    # create the Project to return, add navigator as default integration
    project = arcgis.apps.workforce.Project(workforce_service_item)
    project.integrations.add(integration_id="arcgis-navigator",prompt="Navigate to Assignment",url_template="https://navigator.arcgis.app?stop=${assignment.latitude},${assignment.longitude}&stopname=${assignment.location}&callback=https://workforce.arcgis.app&callbackprompt=Workforce")
    
    return project


def _v1_create_project_item(gis, folder_name, assignments_item, dispatchers_item, workers_item, tracks_item,
                            dispatchers_webmap, workers_webmap, title, summary, folder_id, group_id):
    """
    Creates the project item
    :param gis: An authenticated GIS
    :param folder_name: The name of the folder in which to place the project
    :param assignments_item: The assignments item
    :param dispatchers_item: THe dispatchers item
    :param workers_item: The workers item
    :param tracks_item: The tracks item
    :param dispatchers_webmap: The dispatcher webmap item
    :param workers_webmap: The workers webmap item
    :param title: The title of the project
    :param summary: The summary of the project
    :param folder_id: The folder id
    :param group_id: The group id
    :return: The project item
    """
    item_properties = {
        "title": title,
        "snippet": summary,
        "tags": 'workforce',
        "type": 'Workforce Project',
        "typeKeywords": 'Workforce Project'
    }
    project_data = {
        "workerWebMapId": workers_webmap.id,
        "dispatcherWebMapId": dispatchers_webmap.id,
        "dispatchers": {
            "serviceItemId": dispatchers_item.id,
            "url": dispatchers_item.layers[0].url
        },
        "assignments": {
            "serviceItemId": assignments_item.id,
            "url": assignments_item.layers[0].url
        },
        "workers": {
            "serviceItemId": workers_item.id,
            "url": workers_item.layers[0].url
        },
        "tracks": {
            "serviceItemId": tracks_item.id,
            "url": tracks_item.layers[0].url,
            "enabled": False,
            "updateInterval": 30
        },
        "assignmentIntegrations": [
            {
                "id": 'default-navigator',
                "prompt": "Navigate to Assignment",
                "urlTemplate": "https://navigator.arcgis.app?stop=${assignment.latitude},${assignment.longitude}&stopname=${assignment.location}&callback=https://workforce.arcgis.app&callbackprompt=Workforce",
            }
        ],
        "version": "1.3.0",
        "groupId": group_id,
        "folderId": folder_id
    }
    item_properties["text"] = json.dumps(project_data)
    item = gis.content.add(item_properties, folder=folder_name)
    return item


def _v2_create_worker_webmap(gis, folder_name, workforce_service_item, assignments_popup_def,
                             workers_popup_def, title, summary, thumbnail):
    """
    Creates the worker webmap
    :param gis: An authenticated GIS
    :param folder_name: The name of the folder in which to place the webmap
    :param workforce_service_item: The workforce service item
    :param assignments_popup_def: The assignments popup definition
    :param workers_popup_def: The workers popup definition
    :param title: The title of the project
    :return: The worker webmap item
    """
    extent = _get_default_extent(gis)
    array_extent = [[extent['xmin'], extent['ymin']], [extent['xmax'], extent['ymax']]]
    item_properties = {
        "title": "{}".format(title),
        "snippet": summary,
        "extent": array_extent,
        "type": 'Web Map',
        "typeKeywords": 'ArcGIS Online,Explorer Web Map,Map,Offline,Online Map,Web Map,Workforce Worker,Data Editing',
        "properties": {"workforceFeatureServiceId": workforce_service_item.id}
    }

    webmap_data = {
        "baseMap": _get_basemap(gis),
        "operationalLayers": [],
        "spatialReference": {
            "wkid": 102100,
            "latestWkid": 3857
        },
        "authoringApp": "ArcGISPythonAPI",
        "authoringAppVersion": str(arcgis.__version__),
        "tables": [],
        "applicationProperties": {
            "viewing": {
                "search": {
                    "enabled": True,
                    "disablePlaceFinder": False,
                    "layers": [
                        {
                            "id": "Workers_0",
                            "field": {
                                "name": 'name',
                                "exactMatch": False,
                                "type": 'esriFieldTypeString'
                            }
                        }
                    ]
                }
            }
        },
        "version": "2.11"
    }
    webmap_data["operationalLayers"].append(_build_operational_layers(workforce_service_item, assignments_popup_def, layer_index=0))
    webmap_data["operationalLayers"].append(_build_operational_layers(workforce_service_item, workers_popup_def, layer_index=1))
    webmap_data["tables"].append(_build_table(workforce_service_item, table_index=0))
    webmap_data["tables"].append(_build_table(workforce_service_item, table_index=1))
    webmap_data["tables"].append(_build_table(workforce_service_item, table_index=2))
    item_properties["text"] = json.dumps(webmap_data)

    item = gis.content.add(item_properties, folder=folder_name, thumbnail=thumbnail)
    return item


def _v2_create_dispatcher_webmap(gis, folder_name, workforce_service_item, assignments_popup_def,
                                 workers_popup_def, title, summary, thumbnail):
    """
    Creates the dispatcher webmap
    :param gis: An authenticated GIS
    :param folder_name: The folder in which to store the webmap
    :param workforce_service_item: The workforce service item item
    :param assignments_popup_def: The assignments popup definition
    :param workers_popup_def: The workers popup definition
    :param title: The title of the project
    :return: The dispatcher webmap item
    """
    extent = _get_default_extent(gis)
    array_extent = [[extent['xmin'], extent['ymin']], [extent['xmax'], extent['ymax']]]
    item_properties = {
        "title": "{} Dispatcher Map".format(title),
        "snippet": summary,
        "extent": array_extent,
        "type": 'Web Map',
        "typeKeywords": 'ArcGIS Online,Explorer Web Map,Map,Offline,Online Map,Web Map,Workforce Dispatcher',
        "properties": {"workforceFeatureServiceId": workforce_service_item.id}
    }

    webmap_data = {
        "baseMap": _get_basemap(gis),
        "operationalLayers": [],
        "spatialReference": {
            "wkid": 102100,
            "latestWkid": 3857
        },
        "authoringApp": "ArcGISPythonAPI",
        "authoringAppVersion": str(arcgis.__version__),
        "version": "2.11"
    }
    webmap_data["operationalLayers"].append(_build_operational_layers(workforce_service_item, assignments_popup_def, layer_index=0, capabilities="Query,Sync"))
    webmap_data["operationalLayers"].append(_build_operational_layers(workforce_service_item, workers_popup_def, layer_index=1, capabilities="Query,Sync"))
    item_properties["text"] = json.dumps(webmap_data)
    item = gis.content.add(item_properties, folder=folder_name, thumbnail=thumbnail)
    return item


def _v2_create_service(gis, service_name, folder_name, spatial_ref, item_properties=None):
    """
    Helper method that creates a workforce service
    :param gis: An authenticated GIS
    :param service_name: The name of the service
    :param folder_name: The name of the folder in which to place the service
    :param spatial_ref: The spatial reference to use
    :return: The service item
    """
    create_params = {
        "name": service_name,
        "capabilities": 'Query,Editing,Create,Update,Delete,Sync',
        "spatialReference": spatial_ref,
        "syncRowsMovedOutsideFilter": True,
        "preserveLayerIds": True
    }

    return gis.content.create_service(
        service_name,
        folder=folder_name,
        create_params=create_params,
        item_properties=item_properties,
        service_type="featureService"
    )


def _v2_create_service_with_layers(gis, folder_name, service_name, assignments_layer_def, workers_layer_def, dispatchers_table_def, assignment_type_table_def, integration_table_def, title):
    """
       Creates a service, adds, and layer and optionally enables attachments
       :param gis: An authenticated GIS
       :param folder_name: The name of the folder in which to place the service
       :param service_name: The name of the service
       :param assignments_layer_def: The assignments layer definition (dictionary)
       :param workers_layer_def: The workers layer definition (dictionary)
       :param dispatchers_table_def: The dispatchers table definition (dictionary)
       :param assignment_type_table_def: The assignment type table definition (dictionary)
       :param integration_table_def: The integration table definition (dictionary)
       :return: The service item
    """
    default_extent = gis.properties["defaultExtent"]
    spatial_reference = default_extent["spatialReference"]
    layer_defs = [assignments_layer_def, workers_layer_def]
    table_defs = [dispatchers_table_def, assignment_type_table_def, integration_table_def]
    if gis.content.is_service_name_available(service_name, "featureService"):
        item_properties = {
            "title": title,
            "tags": "workforce",
            "typeKeywords": "Workforce Project"
        }
        item = _v2_create_service(gis, service_name, folder_name, spatial_reference, item_properties)
        feature_layer_collection = arcgis.features.FeatureLayerCollection.fromitem(item)
    else:
        raise Exception("Service name already exists.")

    for layer_def in layer_defs:
        layer_def["extent"] = default_extent

    feature_layer_collection.manager.add_to_definition({
        "layers": layer_defs,
        "tables": table_defs
    })
    
    # enabled editor tracking
    feature_layer_collection.manager.update_definition({
        "editorTrackingInfo": {
            "enableEditorTracking": True,
            "enableOwnershipAccessControl": False,
            "allowOthersToUpdate": True,
            "allowOthersToDelete": True
        }
    })
    return item


def _get_default_extent(gis):
    """
    Helper method to get the extent used by the org or use default extent
    :param gis: An authenticated GIS
    :return: A dictionary representing an extent
    """
    default_extent = gis.properties['defaultExtent']
    try:
        extent = \
        arcgis.geometry.project([arcgis.geometry.Envelope(default_extent)], default_extent["spatialReference"]["wkid"],
                                4326)[0]
    except:
        extent = {
            "spatialReference": {
                "wkid": 4326
            },
            "xmin": -180,
            "ymin": -90,
            "xmax": 180,
            "ymax": 90
        }
    return extent


def _v1_create_project(gis, summary, title):

    for f in gis.users.me.folders:
        if f['title'].lower() == title.lower():
            raise WorkforceError("A folder named '{}' already exists.".format(title))

    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        future_folder = executor.submit(gis.content.create_folder, title)
        future_group = executor.submit(gis.groups.create,
                                       title,
                                       'workforce',
                                       access='private',
                                       is_invitation_only=True,
                                       is_view_only=True,
                                       sort_field='title',
                                       sort_order='asc'
                                       )
    folder = future_folder.result()
    group = future_group.result()
    folder_name = title
    folder_id = folder["id"]
    group_id = group.id
    # create services
    worker_service_name = "workers_{}".format(group_id)
    dispatcher_service_name = "dispatchers_{}".format(group_id)
    assignment_service_name = "assignments_{}".format(group_id)
    tracking_service_name = "location_{}".format(group_id)
    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        workers_item_future = executor.submit(_v1_create_service_with_layer, gis, folder_name, worker_service_name,
                                              worker_layer_definition_v1)
        dispatchers_item_future = executor.submit(_v1_create_service_with_layer, gis, folder_name, dispatcher_service_name,
                                                  dispatcher_layer_definition_v1)
        assignments_item_future = executor.submit(_v1_create_service_with_layer, gis, folder_name, assignment_service_name,
                                                  assignment_layer_definition_v1, attachments=True)
        tracks_item_future = executor.submit(_v1_create_service_with_location_tracking_layer, gis, folder_name,
                                             tracking_service_name, tracking_layer_definition_v1)
    workers_item = workers_item_future.result()
    dispatchers_item = dispatchers_item_future.result()
    assignments_item = assignments_item_future.result()
    tracks_item = tracks_item_future.result()

    # create webmaps
    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        workers_webmap_future = executor.submit(_v1_create_worker_webmap,
                                                gis,
                                                folder_name,
                                                assignments_item,
                                                workers_item,
                                                tracks_item,
                                                assignment_layer_popup_definition_v1,
                                                worker_layer_popup_definition_v1,
                                                title)
        dispatchers_webmap_future = executor.submit(_v1_create_dispatcher_webmap,
                                                    gis,
                                                    folder_name,
                                                    assignments_item,
                                                    workers_item,
                                                    assignment_layer_popup_definition_v1,
                                                    worker_layer_popup_definition_v1,
                                                    title)
    workers_webmap = workers_webmap_future.result()
    dispatchers_webmap = dispatchers_webmap_future.result()

    # create the project
    project_item = _v1_create_project_item(gis,
                                           folder_name,
                                           assignments_item,
                                           dispatchers_item,
                                           workers_item,
                                           tracks_item,
                                           dispatchers_webmap,
                                           workers_webmap,
                                           title,
                                           summary,
                                           folder_id,
                                           group_id,
                                           )
    project_items = [
        workers_item,
        dispatchers_item,
        assignments_item,
        tracks_item,
        workers_webmap,
        dispatchers_webmap,
        project_item
    ]
    # share and protect items
    for i in project_items:
        i.share(groups=[group])
        i.protect()

    my_path = os.path.abspath(os.path.dirname(__file__))
    thumbnail = os.path.join(my_path, '/'.join(('resources', 'default-project-thumbnail.png')))
    project_item.update(thumbnail=thumbnail)

    # manually add the owner as the first dispatcher
    user_id_field_name = "userId"
    for field in dispatchers_item.layers[0].properties.fields:
        if field['name'].lower() == "userid":
            user_id_field_name = field['name']
            break

    dispatchers_item.layers[0].edit_features(adds=[arcgis.features.Feature(
        attributes={
            "name": gis.users.me.fullName,
            user_id_field_name: gis.users.me.username
        }
    )])

    # create the Project to return
    project = arcgis.apps.workforce.Project(project_item)
    return project


def _v1_create_worker_webmap(gis, folder_name, assignments_item, workers_item, tracking_item, assignments_popup_def,
                             workers_popup_def, title):
    """
    Creates the worker webmap
    :param gis: An authenticated GIS
    :param folder_name: The name of the folder in which to place the webmap
    :param assignments_item: The assignments item
    :param workers_item: The workers item
    :param tracking_item: The tracking item
    :param assignments_popup_def: The assignments popup definition
    :param workers_popup_def: The workers popup definition
    :param title: The title of the project
    :return: The worker webmap item
    """
    extent = _get_default_extent(gis)
    array_extent = [[extent['xmin'], extent['ymin']], [extent['xmax'], extent['ymax']]]
    item_properties = {
        "title": "{}_workers".format(title),
        "tags": "workforce-worker",
        "extent": array_extent,
        "type": 'Web Map',
        "typeKeywords": 'ArcGIS Online,Explorer Web Map,Map,Online Map,Web Map,Workforce Project'
    }

    webmap_data = {
        "baseMap": _get_basemap(gis),
        "operationalLayers": [],
        "spatialReference": {
            "wkid": 102100,
            "latestWkid": 3857
        },
        "authoringApp": "ArcGISPythonAPI",
        "authoringAppVersion": str(arcgis.__version__),
        "applicationProperties": {
            "viewing": {
                "search": {
                    "enabled": True,
                    "disablePlaceFinder": False,
                    "layers": [
                        {
                            "id": "Workers_0",
                            "field": {
                                "name": 'name',
                                "exactMatch": False,
                                "type": 'esriFieldTypeString'
                            }
                        }
                    ]
                }
            }
        },
        "version": "2.11"
    }

    webmap_data["operationalLayers"].append(_build_operational_layers(tracking_item, visibility=False, capabilities="Query"))
    webmap_data["operationalLayers"].append(_build_operational_layers(assignments_item, assignments_popup_def, capabilities="Query,Sync"))
    webmap_data["operationalLayers"].append(_build_operational_layers(workers_item, workers_popup_def, capabilities="Query,Sync"))
    item_properties["text"] = json.dumps(webmap_data)

    item = gis.content.add(item_properties, folder=folder_name)
    return item


def _v1_create_dispatcher_webmap(gis, folder_name, assignments_item, workers_item, assignments_popup_def,
                                 workers_popup_def, title):
    """
    Creates the dispatcher webmap
    :param gis: An authenticated GIS
    :param folder_name: The folder in which to store the webmap
    :param assignments_item: The assignments item
    :param workers_item: The workers item
    :param assignments_popup_def: The assignments popup definition
    :param workers_popup_def: The workers popup definition
    :param title: The title of the project
    :return: The dispatcher webmap item
    """
    extent = _get_default_extent(gis)
    array_extent = [[extent['xmin'], extent['ymin']], [extent['xmax'], extent['ymax']]]
    item_properties = {
        "title": "{}_dispatchers".format(title),
        "tags": "workforce-dispatcher",
        "extent": array_extent,
        "type": 'Web Map',
        "typeKeywords": 'ArcGIS Online,Explorer Web Map,Map,Offline,Online Map,Web Map,Workforce Project'
    }

    webmap_data = {
        "baseMap": _get_basemap(gis),
        "operationalLayers": [],
        "spatialReference": {
            "wkid": 102100,
            "latestWkid": 3857
        },
        "authoringApp": "ArcGISPythonAPI",
        "authoringAppVersion": str(arcgis.__version__),
        "version": "2.11"
    }
    webmap_data["operationalLayers"].append(_build_operational_layers(assignments_item, assignments_popup_def, capabilities="Query,Sync"))
    webmap_data["operationalLayers"].append(_build_operational_layers(workers_item, workers_popup_def, capabilities="Query,Sync"))
    item_properties["text"] = json.dumps(webmap_data)
    item = gis.content.add(item_properties, folder=folder_name)
    return item


def _v1_create_service_with_layer(gis, folder_name, service_name, layer_definition, attachments=False, sync=True):
    """
    Creates a service, adds, and layer and optionally enables attachments
    :param gis: An authenticated GIS
    :param folder_name: The name of the folder in which to place the service
    :param service_name: The name of the service
    :param layer_definition: The definition of the layer (dictionary)
    :param attachments: Should attachments be enabled on the service
    :return: The service item
    """
    default_extent = gis.properties["defaultExtent"]
    spatial_reference = default_extent["spatialReference"]
    layer_definition["extent"] = default_extent
    if gis.content.is_service_name_available(service_name, "featureService"):
        item = _v1_create_service(gis, service_name, folder_name, spatial_reference, sync=sync)
        item.update({
            "tags": "workforce"
        })
        feature_layer_collection = arcgis.features.FeatureLayerCollection.fromitem(item)

        # add layer
        feature_layer_collection.manager.add_to_definition({
            "layers": [layer_definition]
        })

        # enabled editor tracking
        feature_layer_collection.manager.update_definition({
            "editorTrackingInfo": {
                "enableEditorTracking": True,
                "enableOwnershipAccessControl": False,
                "allowOthersToUpdate": True,
                "allowOthersToDelete": True
            }
        })

        # enable attachments
        if attachments:
            feature_layer_collection.layers[0].manager.update_definition({
                "hasAttachments": True
            })
        return item
    else:
        raise Exception("Service name already exists.")


def _v1_create_service_with_location_tracking_layer(gis, folder_name, service_name, layer_definition):
    """
    Creates the location tracking service, adds a layer and optionally enables attachments
    :param gis: An authenticated GIS
    :param folder_name: The name of the folder in which to place the service
    :param service_name: The name of the service
    :param layer_definition: The definition of the layer (dictionary)
    :return: The location tracking item
    """
    item = _v1_create_service_with_layer(gis, folder_name, service_name, layer_definition, sync=False)
    item.layers[0].manager.update_definition({
        "timeInfo": {
            "startTimeField": item.layers[0].properties['editFieldsInfo']['creationDateField'],
            "timeReference": {
                "timeZone": 'UTC',
                "respectDaylightSaving": False
            },
            "timeInterval": 0,
            "exportOptions": {
                "useTime": True,
                "timeDataCumulative": True,
                "TimeOffset": 0,
                "timeOffsetUnits": 'esriTimeUnitsCenturies'
            },
            "hasLiveData": True
        }
    })
    item.update({
        "tags": "workforce, Location Tracking",
        "typeKeywords": 'Collector,Data,Feature Service Template,Layer,Layer Template,Location Tracking,Platform,Service template,Template'
    })
    return item


def _v1_create_service(gis, service_name, folder_name, spatial_ref, sync=True):
    """
    Helper method that creates a workforce service
    :param gis: An authenticated GIS
    :param service_name: The name of the service
    :param folder_name: The name of the folder in which to place the service
    :param spatial_ref: The spatial reference to use
    :return: The service item
    """
    create_params = {
        "name": service_name,
        "allowGeometryUpdates": True,
        "units": 'esriMeters',
        "spatialReference": spatial_ref,
        "capabilities": 'Query,Editing,Create,Update,Delete'
    }
    if sync:
        create_params["syncEnabled"] = True
        create_params["syncCapabilities"] = {
            "supportsAsync": True,
            "supportsRegisteringExistingData": True,
            "supportsSyncDirectionControl": True,
            "supportsPerLayerSync": True,
            "supportsPerReplicaSync": True,
            "supportsSyncModelNone": True,
            "supportsRollbackOnFailure": True
        }
        create_params["capabilities"] = 'Query,Editing,Create,Update,Delete,Sync'

    return gis.content.create_service(
        service_name,
        folder=folder_name,
        create_params=create_params,
        service_type="featureService"
    )



import sys
sys.path.insert(0, r"C:\SVN\achapkowski_geosaurus_fork_requests_take2\src")
import pytest
import unittest
from arcgis.gis._impl._con import Connection

class TestConnection(unittest.TestCase):
    pass

if __name__ == "__main__":
    #unittest.main()
    pass



import json
from pathlib import Path
from ._codetemplate import image_classifier_prf
from ._arcgis_model import ArcGISModel
import types
from functools import partial
import logging
logger = logging.getLogger()

try:
    from fastai.basic_train import Learner
    from ._arcgis_model import SaveModelCallback, _resnet_family, _densenet_family, _vgg_family
    from ._unet_utils import is_no_color, predict_batch, show_results_multispectral
    import torch
    from torch import nn
    import torch.nn.functional as F
    from torchvision import models
    from ._unet_utils import LabelCallback
    from ._arcgis_model import _EmptyData
    from fastai.layers import CrossEntropyFlat
    from .._utils.segmentation_loss_functions import  FocalLoss, MixUpCallback, DiceLoss
    from ._psp_utils import PSPNet, _pspnet_learner, _pspnet_learner_with_unet, accuracy
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from .._utils.classified_tiles import per_class_metrics
    import numpy as np
    from fastai.callbacks import EarlyStoppingCallback
    from fastai.torch_core import split_model_idx
    from fastai.vision import flatten_model
    from ._deeplab_utils import compute_miou
    from ._PointRend import PointRend_target_transform
    HAS_FASTAI = True
except Exception as e:
    HAS_FASTAI = False

class PSPNetClassifier(ArcGISModel):

    """
    Model architecture from https://arxiv.org/abs/1612.01105.
    Creates a PSPNet Image Segmentation/ Pixel Classification model. 

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used for
                            creating the base of the `PSPNetClassifier`, which
                            is `resnet50` by default. It supports the ResNet,
                            DenseNet, and VGG families.
    ---------------------   -------------------------------------------
    use_unet                Optional Bool. Specify whether to use Unet-Decoder or not,
                            Default True.                          
    ---------------------   -------------------------------------------
    pyramid_sizes           Optional List. The sizes at which the feature map is pooled at.
                            Currently set to the best set reported in the paper,
                            i.e, (1, 2, 3, 6)
    ---------------------   -------------------------------------------
    pretrained              Optional Bool. If True, use the pretrained backbone                                                                                 
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained PSPNet model is
                            saved.
    ---------------------   -------------------------------------------
    unet_aux_loss           Optional. Bool If True will use auxiliary loss for PSUnet.
                            Default set to False. This flag is applicable only when
                            use_unet is True.
    ---------------------   -------------------------------------------
    pointrend               Optional boolean. If True, it will use PointRend
                            architecture on top of the segmentation head.
                            Default: False. PointRend architecture from
                            https://arxiv.org/pdf/1912.08193.pdf.                        
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    class_balancing         Optional boolean. If True, it will balance the
                            cross-entropy loss inverse to the frequency
                            of pixels per class. Default: False. 
    ---------------------   -------------------------------------------
    mixup                   Optional boolean. If True, it will use mixup
                            augmentation and mixup loss. Default: False
    ---------------------   -------------------------------------------
    focal_loss              Optional boolean. If True, it will use focal loss.
                            Default: False
    ---------------------   -------------------------------------------
    dice_loss_fraction      Optional float. 
                            Min_val=0, Max_val=1 
                            If > 0 , model will use a combination of default or 
                            focal(if focal=True) loss with the specified fraction 
                            of dice loss.
                            E.g. 
                            for dice = 0.3, loss = (1-0.3)*default loss + 0.3*dice
                            Default: 0
    ---------------------   -------------------------------------------    
    ignore_classes          Optional list. It will contain the list of class
                            values on which model will not incur loss.
                            Default: []
    ---------------------   -------------------------------------------
    keep_dilation           Optional boolean. When PointRend architecture is used,
                            keep_dilation=True can potentially improves accuracy
                            at the cost of memory consumption. Default: False
    =====================   ===========================================    

    :returns: `PSPNetClassifier` Object
    """

    def __init__(self, data, backbone=None, use_unet=True, pyramid_sizes=[1, 2, 3, 6], pretrained_path=None, unet_aux_loss=False, pointrend=False, *args, **kwargs):

        # Set default backbone to be 'resnet50'
        if backbone is None: 
            backbone = models.resnet50
        
        self._check_dataset_support(data)
        if not (self._check_backbone_support(backbone)):
            raise Exception (f"Enter only compatible backbones from {', '.join(self.supported_backbones)}")

        super().__init__(data, backbone, **kwargs)

        if pointrend:
            use_unet = False

        self._ignore_classes = kwargs.get('ignore_classes', [])
        if self._ignore_classes != [] and len(data.classes) <= 3:
            raise Exception(f"`ignore_classes` parameter can only be used when the dataset has more than 2 classes.")

        data_classes = list(self._data.class_mapping.keys())
        if 0 not in list(data.class_mapping.values()):
            self._ignore_mapped_class = [data_classes.index(k) + 1 for k in self._ignore_classes if k != 0]
        else:
            self._ignore_mapped_class = [data_classes.index(k) + 1 for k in self._ignore_classes]
        if self._ignore_classes != []:
            if 0 not in self._ignore_mapped_class:
                self._ignore_mapped_class.insert(0, 0)
            global accuracy
            accuracy = partial(accuracy, ignore_mapped_class=self._ignore_mapped_class)
        self.mixup = kwargs.get('mixup', False)
        self.class_balancing = kwargs.get('class_balancing', False)
        self.focal_loss = kwargs.get('focal_loss', False)
        self.dice_loss_fraction = kwargs.get('dice_loss_fraction', False)
        self.weighted_dice = kwargs.get('weighted_dice', False)
        self.keep_dilation = kwargs.get('keep_dilation', False)
        self._vggv2 = kwargs.get('vggv2', True)
        self._code = image_classifier_prf
        self.pyramid_sizes = pyramid_sizes
        self._use_unet = use_unet
        self._unet_aux_loss = unet_aux_loss
        self._pointrend = pointrend

        if use_unet:
            self.learn = _pspnet_learner_with_unet(data,
                                                   backbone=self._backbone,
                                                   chip_size=self._data.chip_size, 
                                                   pyramid_sizes=pyramid_sizes, 
                                                   pretrained=True, 
                                                   metrics=accuracy, 
                                                   unet_aux_loss=unet_aux_loss,
                                                   vggv2=self._vggv2)

            if self.class_balancing and data.class_weight is not None:
                class_weight = torch.tensor([data.class_weight.mean()] + data.class_weight.tolist()).float().to(self._device)
                self.learn.loss_func = CrossEntropyFlat(class_weight, axis=1)

        else:
            self.learn = _pspnet_learner(data, 
                                         backbone=self._backbone, 
                                         chip_size=self._data.chip_size, 
                                         pyramid_sizes=pyramid_sizes, 
                                         pretrained=True,
                                         pointrend=self._pointrend,
                                         keep_dilation=self.keep_dilation,
                                         metrics=accuracy)


        if self.mixup:
            self.learn.callbacks.append(MixUpCallback(self.learn))

        self.learn.callbacks.append(LabelCallback(self.learn))  #appending label callback 

        class_weight = None
        if self.class_balancing:
            if data.class_weight is not None:
                # Handle condition when nodata is already at pixel value 0 in data
                if (data.c - 1) == data.class_weight.shape[0]:
                    class_weight = torch.tensor([data.class_weight.mean()] + data.class_weight.tolist()).float().to(
                        self._device)
                else:
                    class_weight = torch.tensor(data.class_weight).float().to(self._device)
            else:
                if getattr(data, 'overflow_encountered', False):
                    logger.warning("Overflow Encountered. Ignoring `class_balancing` parameter.")
                    class_weight = [1] * len(data.classes)
                else:
                    logger.warning(
                        "Could not find 'NumPixelsPerClass' in 'esri_accumulated_stats.json'. Ignoring `class_balancing` parameter.")

        if self._ignore_classes != []:
            if not self.class_balancing:
                class_weight = torch.tensor([1] * data.c).float().to(self._device)
            class_weight[self._ignore_mapped_class] = 0.

        self.learn.loss_func = CrossEntropyFlat(class_weight, axis=1)
        self._final_class_weight = class_weight


        if unet_aux_loss or not use_unet:
            self.learn.loss_func = self._psp_loss          
        if self.focal_loss:
            self.learn.loss_func = FocalLoss(self.learn.loss_func)            
        if self.dice_loss_fraction:
            self.learn.loss_func = DiceLoss(self.learn.loss_func, self.dice_loss_fraction,  weighted_dice=self.weighted_dice)
        self.learn.model = self.learn.model.to(self._device)
        self.freeze()
        self._arcgis_init_callback() # make first conv weights learnable

        if pretrained_path is not None:
            self.load(pretrained_path)


    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    # Return a list of supported backbones names
    @property
    def supported_backbones(self):
        """ Supported torchvision backbones for this model. """        
        return PSPNetClassifier._supported_backbones()

    @staticmethod
    def _supported_backbones():
        return [*_resnet_family, *_densenet_family, *_vgg_family]

    @property
    def  supported_datasets(self):
        """ Supported dataset types for this model. """
        return PSPNetClassifier._supported_datasets()
    
    @staticmethod
    def _supported_datasets():
        return ['Classified_Tiles']    
    
    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a PSPNet classifier from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :returns: `PSPNetClassifier` Object
        """
        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)
            
        model_file = Path(emd['ModelFile'])
        
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file
            
        model_params = emd['ModelParameters']
        model_params['vggv2'] = model_params.get('vggv2', False)

        try:
            class_mapping = {i['Value'] : i['Name'] for i in emd['Classes']}
            color_mapping = {i['Value'] : i['Color'] for i in emd['Classes']}
        except KeyError:
            class_mapping = {i['ClassValue'] : i['ClassName'] for i in emd['Classes']} 
            color_mapping = {i['ClassValue'] : i['Color'] for i in emd['Classes']}                

        if data is None:
            data = _EmptyData(path=emd_path.parent.parent, loss_func=None, c=len(class_mapping) + 1, chip_size=emd['ImageHeight'])
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data = get_multispectral_data_params_from_emd(data, emd)
            data.emd_path = emd_path
            data.emd = emd

        return cls(data, **model_params, pretrained_path=str(model_file))

    def _psp_loss(self, outputs, targets, **kwargs):
        targets = targets.squeeze(1).detach()
        
        criterion = nn.CrossEntropyLoss(weight=self._final_class_weight).to(self._device)
        if self.learn.model.training:
            out = outputs[0]
            aux = outputs[1]
            if self._pointrend:
                pointrend_out = outputs[2][0]
                pointrend_coord = outputs[2][1]
                pointrend_target = PointRend_target_transform(targets, pointrend_coord)
        else: # validation
            out = outputs
        main_loss = criterion(out, targets)

        if self.learn.model.training:
            aux_loss = criterion(aux, targets)
            if self._pointrend:
                pointrend_loss = criterion(pointrend_out, pointrend_target)
                total_loss = main_loss + 0.4 * aux_loss + pointrend_loss
            else:
                total_loss = main_loss + 0.4 * aux_loss
            return total_loss
        else:
            return main_loss

    def freeze(self):
        "Freezes the pretrained backbone."
        for idx, i in enumerate(flatten_model(self.learn.model)):
            if hasattr(i, 'dilation'):
                if isinstance(i, (nn.BatchNorm2d)):
                    continue
                dilation = i.dilation
                dilation = dilation[0] if isinstance(dilation, tuple) else dilation
                if dilation > 1:
                    break        
            for p in i.parameters():
                p.requires_grad = False

        self.learn.layer_groups = split_model_idx(self.learn.model, [idx])  ## Could also call self.learn.freeze after this line because layer groups are now present.      
        self.learn.create_opt(lr=3e-3)

    def unfreeze(self):
        """
        Unfreezes the earlier layers of the model for fine-tuning.
        """
        for _, param in self.learn.model.named_parameters():
            param.requires_grad = True

    def accuracy(self, input=None, target=None, void_code=0, class_mapping=None):
        if input is not None or target is not None:
            accuracy(input, target)
        else:
            try:
                return self.learn.validate()[1].tolist()
            except Exception as e:
                accuracy = self._data.emd.get('accuracy')
                if accuracy:
                    return accuracy
                else:
                    logger.error("Metric not found in the loaded model")

        
    def _get_emd_params(self, save_inference_file):
        import random
        _emd_template = {"ModelParameters" : {}}
        _emd_template["ModelType"] = "ImageClassification"
        _emd_template["ModelParameters"]["pyramid_sizes"] = self.pyramid_sizes
        _emd_template["ModelParameters"]["use_unet"] = self._use_unet
        _emd_template["ModelParameters"]["pointrend"] = self._pointrend
        _emd_template["ModelParameters"]["keep_dilation"] = self.keep_dilation
        _emd_template["ModelParameters"]["vggv2"] = self._vggv2
        _emd_template["ModelParameters"]["unet_aux_loss"] = self._unet_aux_loss
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_psp"
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISImageClassifier.py"
        else:
            _emd_template["InferenceFunction"] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISImageClassifier.py"
        _emd_template["ExtractBands"] = [0, 1, 2]
        _emd_template["ignore_mapped_class"] = self._ignore_mapped_class
        _emd_template['Classes'] = []
        class_data = {}
        for i, class_name in enumerate(self._data.classes[1:]):  # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)] if is_no_color(self._data.color_mapping) else \
            self._data.color_mapping[inverse_class_mapping[class_name]]
            class_data["Color"] = color
            _emd_template['Classes'].append(class_data.copy())

        return _emd_template

    def show_results(self, rows=5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.
        """
        self._check_requisites()
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)
        self.learn.show_results(rows=rows, ignore_mapped_class=self._ignore_mapped_class, **kwargs)

    def _show_results_multispectral(self, rows=5, alpha=0.7, **kwargs): # parameters adjusted in kwargs
        return_fig = kwargs.get('return_fig', False)
        ret_val = show_results_multispectral(
            self, 
            nrows=rows, 
            alpha=alpha, 
            **kwargs
        )
        if return_fig:
            fig, ax = ret_val
            return fig

    @property
    def _model_metrics(self):
        return {'accuracy': '{0:1.4e}'.format(self._get_model_metrics())}

    def _get_model_metrics(self, **kwargs):
        checkpoint = kwargs.get('checkpoint', True)
        if not hasattr(self.learn, 'recorder'):
            return 0.0

        try:
            model_accuracy = self.learn.recorder.metrics[-1][0]
            if checkpoint:
                val_losses = self.learn.recorder.val_losses
                model_accuracy = self.learn.recorder.metrics[val_losses.index(min(val_losses))][0]
        except:
            model_accuracy = 0.0

        return float(model_accuracy)

    def mIOU(self, mean=False, show_progress=True):

        """
        Computes mean IOU on the validation set for each class.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                mean IOU, otherwise returns mean iou of all
                                classes combined.   
        ---------------------   -------------------------------------------
        show_progress           Optional bool. Displays the progress bar if
                                True.                     
        =====================   ===========================================
        
        :returns: `dict` if mean is False otherwise `float`
        """
        self._check_requisites()
        num_classes = torch.arange(self._data.c)
        miou = compute_miou(self, self._data.valid_dl, mean, num_classes, show_progress, self._ignore_mapped_class)
        if mean:
            miou = [miou[i] for i in range(len(miou)) if i not in self._ignore_mapped_class]
            return np.mean(miou)
        if self._ignore_mapped_class == []:
            return dict(zip(['0'] + self._data.classes[1:], miou))
        else:
            class_values = [0] + list(self._data.class_mapping.keys())
            return {class_values[i]: miou[i] for i in range(len(miou)) if i not in self._ignore_mapped_class} 


    def per_class_metrics(self, ignore_classes=[]):
        """
        Computer per class precision, recall and f1-score on validation set.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        self                    segmentation model object -> [PSPNetClassifier | UnetClassifier | DeepLab]
        ---------------------   -------------------------------------------
        ignore_classes          Optional list. It will contain the list of class
                                values on which model will not incur loss.
                                Default: []    
        =====================   ===========================================

        Returns per class precision, recall and f1 scores 
        """
        ## Calling imported function `per_class_metrics`        
        try:
            self._check_requisites()
            ## Calling imported function `per_class_metrics`
            return per_class_metrics(self, ignore_classes)
        except:
            import pandas as pd
            return pd.read_json(self._data.emd['per_class_metrics'])



#!/usr/bin/env python3


"""
The Ramer-Douglas-Peucker algorithm roughly ported from the pseudo-code provided
by http://en.wikipedia.org/wiki/Ramer-Douglas-Peucker_algorithm

The code is taken from
https://github.com/mitroadmaps/roadtracer/blob/master/lib/discoverlib/rdp.py

License:
================================================================================
Copyright (c) 2018 Favyen Bastani
Copyright (c) 2018 Massachusetts Institute of Technology

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from math import sqrt


def distance(a, b):
    return sqrt((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2)


def point_line_distance(point, start, end):
    """
    Calaculate the prependicuar distance of given point from the line having
    start and end points.
    """
    if start == end:
        return distance(point, start)
    else:
        n = abs(
            (end[0] - start[0]) * (start[1] - point[1])
            - (start[0] - point[0]) * (end[1] - start[1])
        )
        d = sqrt((end[0] - start[0]) ** 2 + (end[1] - start[1]) ** 2)
        return n / d


def rdp(points, epsilon):
    """
    Reduces a series of points to a simplified version that loses detail, but
    maintains the general shape of the series.

    @param points: Series of points for a line geometry represnted in graph.
    @param epsilon: Tolerance required for RDP algorithm to aproximate the
                    line geometry.

    @return: Aproximate series of points for approximate line geometry
    """
    dmax = 0.0
    index = 0
    for i in range(1, len(points) - 1):
        d = point_line_distance(points[i], points[0], points[-1])
        if d > dmax:
            index = i
            dmax = d
    if dmax >= epsilon:
        results = rdp(points[: index + 1], epsilon)[:-1] + rdp(points[index:], epsilon)
    else:
        results = [points[0], points[-1]]
    return results



# Referred from https://github.com/tcapelle/timeseries_fastai
import torch
import torch.nn as nn
from .layers import *


class _TSFCN(nn.Module):
    def __init__(self, input, output, layers=[128, 256, 128], kss=[7, 5, 3], device=None):
        super().__init__()

        if not isinstance(kss, list) or len(kss) == 0:
            kss = [7, 5, 3]

        i = 0

        self.conv_layers = nn.ModuleList([])
        sizes = zip([input] + layers, layers)
        for n1, n2 in sizes:
            if i < len(kss):
                ks = kss[i]
            else:
                ks = kss[-1]
            self.conv_layers.append(convlayer(n1, n2, ks))

        self.avg_pooling = nn.AdaptiveAvgPool1d(1)
        self.linear = nn.Linear(layers[-1], output)

    def forward(self, x):
        for layer in self.conv_layers:
            x = layer(x)

        x = self.avg_pooling(x).squeeze(-1)
        x = self.linear(x)
        return x
